module mpi_include
  use mpi_typedefs ! derived types and overloaded operators
  implicit none
!  private
  include 'mpif.h'

  public :: fmpi_def_init

  ! Our own special variables
  type(mpi_comm_t), public :: fmpi_comm=mpi_comm_t(-1)  ! Current communicator
  integer, public, parameter :: fmpi_root_rank = 0 
  integer, public :: fmpi_size=-1
  logical, public :: fmpi_root=.false.
  integer, public :: fmpi_rank=-1

  Integer, parameter, public :: FMPI_MAX_PROCESSOR_NAME = MPI_MAX_PROCESSOR_NAME
  Integer, parameter, public :: FMPI_MAX_ERROR_STRING   = MPI_MAX_ERROR_STRING

  Integer, public :: FMPI_STATUS_SIZE
  Integer, public :: FMPI_SOURCE, FMPI_TAG, FMPI_ERROR
  type(mpi_ierr_t), public :: FMPI_SUCCESS, FMPI_ERR_BUFFER, FMPI_ERR_COUNT, &
       FMPI_ERR_TYPE, FMPI_ERR_TAG, FMPI_ERR_COMM, FMPI_ERR_RANK, &
       FMPI_ERR_ROOT, FMPI_ERR_GROUP, FMPI_ERR_OP, FMPI_ERR_TOPOLOGY, &
       FMPI_ERR_DIMS, FMPI_ERR_ARG, FMPI_ERR_UNKNOWN, FMPI_ERR_TRUNCATE, &
       FMPI_ERR_OTHER, FMPI_ERR_INTERN, FMPI_ERR_IN_STATUS, &
       FMPI_ERR_REQUEST, FMPI_ERR_LASTCODE 
  Integer, public :: FMPI_UNDEFINED
  Integer, public :: FMPI_GRAPH, FMPI_CART
  Integer, public :: FMPI_PROC_NULL 
  Integer, public :: FMPI_BSEND_OVERHEAD
  type(mpi_comm_t), public :: FMPI_COMM_NULL
  type(mpi_datatype_t), public :: FMPI_DATATYPE_NULL 
  type(mpi_errhandler_t), public :: FMPI_ERRHANDLER_NULL
  type(mpi_group_t), public :: FMPI_GROUP_NULL 
  Integer, public :: FMPI_KEYVAL_INVALID 
  type(mpi_errhandler_t), public :: FMPI_ERRORS_ARE_FATAL, FMPI_ERRORS_RETURN 
  type(mpi_datatype_t), public ::  FMPI_COMPLEX, FMPI_DOUBLE_COMPLEX,&
       FMPI_LOGICAL, FMPI_REAL, FMPI_DOUBLE_PRECISION, FMPI_INTEGER,&
       FMPI_2INTEGER, FMPI_2COMPLEX, FMPI_2DOUBLE_COMPLEX, FMPI_2REAL, &
       FMPI_2DOUBLE_PRECISION, FMPI_CHARACTER, FMPI_BYTE, FMPI_UB, &
       FMPI_LB, FMPI_PACKED
  type(mpi_datatype_t), public :: FMPI_INTEGER1, FMPI_INTEGER2, FMPI_INTEGER4
  type(mpi_datatype_t), public :: FMPI_REAL4, FMPI_REAL8
  type(mpi_op_t), public :: FMPI_MAX, FMPI_MIN, FMPI_SUM, FMPI_PROD, &
       FMPI_LAND, FMPI_BAND, FMPI_LOR, FMPI_BOR, FMPI_LXOR, &
       FMPI_BXOR, FMPI_MINLOC, FMPI_MAXLOC, FMPI_OP_NULL
  integer, public :: FMPI_GROUP_EMPTY, FMPI_COMM_SELF, FMPI_TAG_UB, &
       FMPI_HOST, FMPI_IO, FMPI_WTIME_IS_GLOBAL6
  type(mpi_comm_t), public :: FMPI_COMM_WORLD
  Integer, public :: FMPI_ANY_SOURCE
  type(mpi_tag_t), public :: FMPI_ANY_TAG 
  integer, public :: FMPI_IDENT, FMPI_CONGRUENT, FMPI_SIMILAR, FMPI_UNEQUAL
  type(mpi_request_t), public ::  FMPI_REQUEST_NULL
  integer, public :: FMPI_WTIME_IS_GLOBAL
contains
  subroutine fmpi_def_init()

    FMPI_STATUS_SIZE=MPI_STATUS_SIZE
    FMPI_SOURCE=MPI_SOURCE
    FMPI_TAG=MPI_TAG
    FMPI_ERROR=MPI_ERROR

    FMPI_SUCCESS = mpi_ierr_t(MPI_SUCCESS)
    FMPI_ERR_BUFFER = mpi_ierr_t(MPI_ERR_BUFFER)
    FMPI_ERR_COUNT = mpi_ierr_t(MPI_ERR_COUNT)
    FMPI_ERR_TYPE = mpi_ierr_t(MPI_ERR_TYPE)
    FMPI_ERR_TAG = mpi_ierr_t(MPI_ERR_TAG)
    FMPI_ERR_COMM = mpi_ierr_t(MPI_ERR_COMM)
    FMPI_ERR_RANK = mpi_ierr_t(MPI_ERR_RANK)
    FMPI_ERR_ROOT = mpi_ierr_t(MPI_ERR_ROOT)
    FMPI_ERR_GROUP = mpi_ierr_t(MPI_ERR_GROUP)
    FMPI_ERR_OP = mpi_ierr_t(MPI_ERR_OP)
    FMPI_ERR_TOPOLOGY = mpi_ierr_t(MPI_ERR_TOPOLOGY)
    FMPI_ERR_DIMS = mpi_ierr_t(MPI_ERR_DIMS)
    FMPI_ERR_ARG = mpi_ierr_t(MPI_ERR_ARG)
    FMPI_ERR_UNKNOWN = mpi_ierr_t(MPI_ERR_UNKNOWN)
    FMPI_ERR_TRUNCATE = mpi_ierr_t(MPI_ERR_TRUNCATE)
    FMPI_ERR_OTHER = mpi_ierr_t(MPI_ERR_OTHER)
    FMPI_ERR_INTERN = mpi_ierr_t(MPI_ERR_INTERN)
    FMPI_ERR_IN_STATUS = mpi_ierr_t(MPI_ERR_IN_STATUS)
    FMPI_ERR_REQUEST = mpi_ierr_t(MPI_ERR_REQUEST)
    FMPI_ERR_LASTCODE = mpi_ierr_t(MPI_ERR_LASTCODE) 
    !
    FMPI_UNDEFINED = MPI_UNDEFINED
    !
    FMPI_GRAPH = MPI_GRAPH
    FMPI_CART  = MPI_CART 
    FMPI_PROC_NULL = MPI_PROC_NULL
    !
    FMPI_BSEND_OVERHEAD  = MPI_BSEND_OVERHEAD 

     !
    FMPI_COMM_NULL = mpi_comm_t(MPI_COMM_NULL)
    !
    FMPI_DATATYPE_NULL = mpi_datatype_t(MPI_DATATYPE_NULL) 

    FMPI_ERRHANDLER_NULL = mpi_errhandler_t(MPI_ERRHANDLER_NULL)

    FMPI_GROUP_NULL = mpi_group_t(MPI_GROUP_NULL)

    FMPI_KEYVAL_INVALID  = MPI_KEYVAL_INVALID 

    FMPI_REQUEST_NULL = mpi_request_t(MPI_REQUEST_NULL)
    ! 
    FMPI_IDENT=MPI_IDENT
    FMPI_CONGRUENT=MPI_CONGRUENT
    FMPI_SIMILAR = MPI_SIMILAR
    FMPI_UNEQUAL = MPI_UNEQUAL

    FMPI_ERRORS_ARE_FATAL = mpi_errhandler_t(MPI_ERRORS_ARE_FATAL)
    FMPI_ERRORS_RETURN = mpi_errhandler_t(MPI_ERRORS_RETURN)
    !

    FMPI_COMPLEX = mpi_datatype_t(MPI_COMPLEX) 
    FMPI_DOUBLE_COMPLEX = mpi_datatype_t(MPI_DOUBLE_COMPLEX)
    FMPI_LOGICAL = mpi_datatype_t(MPI_LOGICAL) 
    FMPI_REAL = mpi_datatype_t(MPI_REAL) 
    FMPI_DOUBLE_PRECISION = mpi_datatype_t(MPI_DOUBLE_PRECISION) 
    FMPI_INTEGER = mpi_datatype_t(MPI_INTEGER)
    FMPI_2INTEGER = mpi_datatype_t(MPI_2INTEGER)
    FMPI_2COMPLEX = mpi_datatype_t(MPI_2COMPLEX)
    FMPI_2DOUBLE_COMPLEX = mpi_datatype_t(MPI_2DOUBLE_COMPLEX)
    FMPI_2REAL = mpi_datatype_t(MPI_2REAL)
    FMPI_2DOUBLE_PRECISION = mpi_datatype_t(MPI_2DOUBLE_PRECISION)
    FMPI_CHARACTER = mpi_datatype_t(MPI_CHARACTER)
    FMPI_BYTE = mpi_datatype_t(MPI_BYTE)
    FMPI_UB = mpi_datatype_t(MPI_UB)
    FMPI_LB = mpi_datatype_t(MPI_LB)
    FMPI_PACKED = mpi_datatype_t(MPI_PACKED)

    FMPI_INTEGER1 = mpi_datatype_t(MPI_INTEGER1)
    FMPI_INTEGER2 = mpi_datatype_t(MPI_INTEGER2)
    FMPI_INTEGER4 = mpi_datatype_t(MPI_INTEGER4)

    FMPI_REAL4 = mpi_datatype_t(MPI_REAL4)
    FMPI_REAL8 = mpi_datatype_t(MPI_REAL8)

    FMPI_MAX = mpi_op_t(MPI_MAX)
    FMPI_MIN = mpi_op_t(MPI_MIN)
    FMPI_SUM = mpi_op_t(MPI_SUM)
    FMPI_PROD = mpi_op_t(MPI_PROD)
    FMPI_LAND = mpi_op_t(MPI_LAND)
    FMPI_BAND = mpi_op_t(MPI_BAND)
    FMPI_LOR = mpi_op_t(MPI_LOR)
    FMPI_BOR = mpi_op_t(MPI_BOR)
    FMPI_LXOR = mpi_op_t(MPI_LXOR)
    FMPI_BXOR = mpi_op_t(MPI_BXOR)
    FMPI_MINLOC = mpi_op_t(MPI_MINLOC)
    FMPI_MAXLOC = mpi_op_t(MPI_MAXLOC)
    FMPI_OP_NULL = mpi_op_t(MPI_OP_NULL)
    !
    FMPI_GROUP_EMPTY = MPI_GROUP_EMPTY
    FMPI_COMM_SELF = MPI_COMM_SELF
    FMPI_TAG_UB = MPI_TAG_UB
    FMPI_HOST = MPI_HOST
    FMPI_IO = MPI_IO
    FMPI_WTIME_IS_GLOBAL = MPI_WTIME_IS_GLOBAL
    FMPI_COMM_WORLD = mpi_comm_t(MPI_COMM_WORLD)
    !
    FMPI_ANY_SOURCE = MPI_ANY_SOURCE
    FMPI_ANY_TAG = mpi_tag_t(MPI_ANY_TAG)
  end subroutine fmpi_def_init
end module mpi_include
