  ! rewritten 6/3/98 by Dr. Jim Mcelwaine
  ! designed to use optional arguments to eliminate as many arguments as possible
  ! and to provide interfaces for user defined types
  ! all types are overloaded for scalars and arrays
  ! in the case of arrays size is used to calculate the number of elements
  ! types are as follwos
  ! mpi type                  fortran type
  ! mpi_integer               integer
  ! mpi_real                  real
  ! mpi_double_precision      double precision
  ! mpi_complex               complex
  ! mpi_logical               logical
  ! mpi_character             character

module mpi_main
  use mpi_include 
  use mpi_typedefs 
  implicit none


contains
  ! Special functions
  function fmpi_is_root(rank)
    integer, intent(in), optional :: rank
    logical :: fmpi_is_root
    if (present(rank)) then 
       fmpi_is_root = (rank==fmpi_root_rank)
    else 
       fmpi_is_root = (fmpi_rank==fmpi_root_rank)
    end if
  end function fmpi_is_root

  ! Set the default communicator, size, rank and root
  subroutine fmpi_set_comm(comm)
    type(mpi_comm_t), intent(in) :: comm
    fmpi_comm = comm
    call fmpi_comm_rank(fmpi_rank)
    call fmpi_comm_size(fmpi_size)
    fmpi_root = fmpi_is_root()
  end subroutine fmpi_set_comm

  ! Mpi interfaces
  function fmpi_wtime() 
    double precision :: fmpi_wtime 
    interface  
       function mpi_wtime() 
         double precision :: mpi_wtime 
       end function mpi_wtime
    end interface
    fmpi_wtime = mpi_wtime()
  end function fmpi_wtime

  function fmpi_wtick() 
    double precision :: fmpi_wtick 
    interface  
       function mpi_wtick() 
         double precision :: mpi_wtick 
       end function mpi_wtick
    end interface
    fmpi_wtick = mpi_wtick()
  end function fmpi_wtick

  subroutine fmpi_finalize(ierror) ! p. 292
    type(mpi_ierr_t), intent(out), optional :: ierror  
    type(mpi_ierr_t) :: ierror_o
    external mpi_finalize
    call mpi_finalize(ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_finalize

  subroutine fmpi_initialized(flag, ierror)
    logical, intent(out) :: flag 
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_initialized
    call mpi_initialized(flag, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_initialized

  subroutine fmpi_abort(errorcode, comm, ierror)
    integer, intent(in), optional  :: errorcode
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    integer :: errorcode_o=0
    type(mpi_ierr_t) :: ierror_o 
    external mpi_abort
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    if (present(errorcode)) errorcode_o = errorcode
    call mpi_abort(comm_o, errorcode_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_abort

  ! Section 2.8.5 Freeing requests
  subroutine fmpi_request_free(request, ierror) ! p. 59
    type(mpi_request_t), intent(inout) :: request
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_request_free
    call mpi_request_free(request, ierror_o)
    if (present(ierror))  ierror = ierror_o
  end subroutine fmpi_request_free

  ! 4 mpi_wait functions
  subroutine fmpi_wait(request, status, ierror)
    type(mpi_request_t), intent(inout) :: request
    type(mpi_status_t), intent(out), optional :: status
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    type(mpi_status_t) :: status_o
    external mpi_wait
    call mpi_wait(request, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_wait
  subroutine fmpi_waitany(request, index, status, ierror)
    type(mpi_request_t), intent(inout) :: request(:)
    integer, intent(out) :: index
    type(mpi_status_t), intent(out), optional :: status
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    type(mpi_status_t) :: status_o
    external mpi_waitany
    call mpi_waitany(size(request),request, index, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_waitany
  subroutine fmpi_waitall(request, status, ierror)
    type(mpi_request_t), intent(inout) :: request(:)
    type(mpi_status_t), intent(out), optional :: status(size(request))
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    type(mpi_status_t) :: status_o(size(request))
    external mpi_waitall
    call mpi_waitall(size(request),request, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_waitall
  subroutine fmpi_waitsome(request, outcount, index, status, ierror)
    type(mpi_request_t), intent(inout) :: request(:)
    integer, intent(out) :: outcount
    integer, intent(out) :: index(size(request))
    type(mpi_status_t), intent(out), optional :: status(size(request))
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    type(mpi_status_t)  :: status_o(size(request))
    external mpi_waitsome
    call mpi_waitsome(size(request),request, outcount, index, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_waitsome

  ! 4 mpi_test functions
  subroutine fmpi_test(request, flag, status, ierror)  ! p.53
    type(mpi_request_t), intent(inout) :: request
    logical, intent(out) :: flag
    type(mpi_status_t), intent(out), optional :: status
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    type(mpi_status_t) :: status_o
    external mpi_test
    call mpi_test(request, flag, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_test
  subroutine fmpi_testany(request, index, flag, status, ierror) !p. 68
    type(mpi_request_t), intent(inout) :: request(:)
    integer, intent(out) :: index
    logical, intent(out) :: flag
    type(mpi_status_t), intent(out), optional :: status
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    type(mpi_status_t) :: status_o
    external mpi_testany
    call mpi_testany(size(request),request, index, flag, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_testany
  subroutine fmpi_testall(request, flag, status, ierror) ! p. 71
    type(mpi_request_t), intent(inout) :: request(:)
    logical, intent(out) :: flag
    type(mpi_status_t), intent(out), optional :: status(size(request))
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    type(mpi_status_t) :: status_o(size(request))
    external mpi_testall
    call mpi_testall(size(request), request, flag, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_testall
  subroutine fmpi_testsome(request, outcount, index, status, ierror) ! p.73
    type(mpi_request_t), intent(inout) :: request(:)
    integer, intent(out) :: outcount
    integer, intent(out) :: index(size(request))
    type(mpi_status_t), intent(out), optional :: status(size(request))
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    type(mpi_status_t)  :: status_o(size(request))
    external mpi_testsome
    call mpi_testsome(size(request),request, outcount, index, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_testsome

  ! Probeing and cancelling commands 2.10 p.75-
  subroutine fmpi_iprobe(source, tag, flag, status, comm, ierror)
    integer, intent(in)  :: source
    type(mpi_tag_t), intent(in)  :: tag
    logical, intent(out) :: flag 
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_status_t), intent(out), optional :: status
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_status_t) :: status_o
    type(mpi_ierr_t) :: ierror_o
    external  mpi_iprobe
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_iprobe (source, tag, comm_o, flag, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_iprobe

  subroutine fmpi_probe(source, tag, status, comm, ierror)
    integer, intent(in)  :: source
    type(mpi_tag_t), intent(in)  :: tag
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_status_t), intent(out), optional :: status
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_status_t) :: status_o
    type(mpi_ierr_t) :: ierror_o
    external  mpi_probe
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_probe (source, tag, comm_o, status_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
    if (present(status)) status = status_o
  end subroutine fmpi_probe 

  subroutine fmpi_cancel(request, ierror)  ! p. 79
    type(mpi_request_t), intent(in)  :: request
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external  mpi_cancel
    call mpi_cancel(request, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cancel

  subroutine fmpi_test_cancelled(status, flag, ierror) ! p.80
    type(mpi_status_t), intent(in) :: status
    logical, intent(out) :: flag 
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external  mpi_test_cancelled
    call mpi_test_cancelled(status, flag, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_test_cancelled

  ! Section 2.11 p. 81 Persistent Communication requests
  ! mpi_send_init is choice function
  ! mpi_recv_init is choice function

  subroutine fmpi_start(request, ierror) ! p. 82
    type(mpi_request_t), intent(inout) :: request
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_start
    call mpi_start(request, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_start

  subroutine fmpi_startall(request, ierror)  ! p. 83
    type(mpi_request_t), intent(inout) :: request(:)
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_startall
    call mpi_startall(size(request), request, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_startall

  ! Datatype Constructors
  subroutine fmpi_type_contiguous(count, oldtype, newtype, ierror)
    integer, intent(in)  :: count
    type(mpi_datatype_t), intent(in)  :: oldtype
    type(mpi_datatype_t), intent(out) :: newtype
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_contiguous
    call mpi_type_contiguous(count, oldtype, newtype, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_contiguous

  subroutine fmpi_type_vector(count, blocklength, stride, oldtype, newtype, ierror)
    integer, intent(in)  :: count
    integer, intent(in)  :: blocklength
    integer, intent(in)  :: stride
    type(mpi_datatype_t), intent(in)  :: oldtype
    type(mpi_datatype_t), intent(out) :: newtype
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_vector
    call mpi_type_vector(count, blocklength, stride, oldtype, newtype, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_vector

  subroutine fmpi_type_hvector(count, blocklength, stride, oldtype, newtype, ierror)
    integer, intent(in)  :: count
    integer, intent(in)  :: blocklength
    integer, intent(in)  :: stride
    type(mpi_datatype_t), intent(in)  :: oldtype
    type(mpi_datatype_t), intent(out) :: newtype
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_hvector
    call mpi_type_hvector(count, blocklength, stride, oldtype, newtype, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_hvector

  subroutine fmpi_type_indexed(blocklength, displacement, oldtype, newtype, ierror)
    integer, intent(in)  :: blocklength(:)
    integer, intent(in)  :: displacement(size(blocklength))
    type(mpi_datatype_t), intent(in)  :: oldtype
    type(mpi_datatype_t), intent(out) :: newtype
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_indexed
    call mpi_type_indexed(size(blocklength),blocklength,displacement,&
         oldtype,newtype,ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_indexed

  subroutine fmpi_type_hindexed(blocklength, displacement, oldtype, newtype, ierror)
    integer, intent(in)  :: blocklength(:)
    integer, intent(in)  :: displacement(size(blocklength))
    type(mpi_datatype_t), intent(in)  :: oldtype
    type(mpi_datatype_t), intent(out) :: newtype
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_hindexed
    call mpi_type_hindexed(size(blocklength),blocklength,displacement,&
         oldtype,newtype,ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_hindexed

  subroutine fmpi_type_struct(blocklength, displacement, oldtype, newtype, ierror)
    integer, intent(in)  :: blocklength(:)
    integer, intent(in)  :: displacement(size(blocklength))
    type(mpi_datatype_t), intent(in)  :: oldtype(size(blocklength))
    type(mpi_datatype_t), intent(out) :: newtype
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_struct
    call  mpi_type_struct (size(blocklength),blocklength,displacement,&
         oldtype,newtype,ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_struct

  subroutine fmpi_type_extent(datatype, extent, ierror)
    type(mpi_datatype_t), intent(in)  :: datatype
    integer, intent(out) :: extent
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_extent
    call mpi_type_extent(datatype, extent, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_extent

  subroutine fmpi_type_size(datatype, ssize, ierror)
    type(mpi_datatype_t), intent(in)  :: datatype
    integer, intent(out) :: ssize
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_size
    call mpi_type_size(datatype, ssize, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_size

  subroutine fmpi_type_lb( datatype, displacement, ierror)
    type(mpi_datatype_t), intent(in)  :: datatype
    integer, intent(out) :: displacement
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_lb
    call mpi_type_lb( datatype, displacement, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_lb

  subroutine fmpi_type_ub( datatype, displacement, ierror)
    type(mpi_datatype_t), intent(in)  :: datatype
    integer, intent(out) :: displacement
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_ub
    call mpi_type_ub( datatype, displacement, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_ub

  subroutine fmpi_type_commit(datatype, ierror)
    type(mpi_datatype_t), intent(inout) :: datatype
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_commit
    call mpi_type_commit(datatype, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_commit

  subroutine fmpi_type_free(datatype, ierror)
    type(mpi_datatype_t), intent(inout) :: datatype
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_type_free
    call mpi_type_free(datatype, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_type_free

  subroutine fmpi_get_elements(status, datatype, count, ierror)
    type(mpi_status_t), intent(in)  :: status
    type(mpi_datatype_t), intent(in)  :: datatype
    integer, intent(out) :: count
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_get_elements
    call mpi_get_elements(status, datatype, count, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_get_elements

  subroutine fmpi_pack_size(incount, datatype, ssize, comm, ierror)
    integer, intent(in)  :: incount
    type(mpi_datatype_t), intent(in)  :: datatype
    integer, intent(out) :: ssize
    type(mpi_comm_t), intent(in),optional :: comm 
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o 
    external mpi_pack_size
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_pack_size(incount, datatype, ssize, comm_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_pack_size


  subroutine fmpi_barrier(comm, ierror) 
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o 
    external    mpi_barrier
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call  mpi_barrier(comm_o, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_barrier

  subroutine fmpi_op_create(function, commute, op, ierror) 
    external function 
    logical, intent(in)  :: commute 
    type(mpi_op_t), intent(out) :: op
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o 
    external   mpi_op_create 
    call  mpi_op_create(function, commute, op, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_op_create

  subroutine fmpi_op_free(op, ierror) 
    type(mpi_op_t), intent(in)  :: op
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o 
    external    mpi_op_free
    call mpi_op_free(op, ierror_o )  
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_op_free
  subroutine fmpi_get_processor_name(name, resultlen, ierror)
    character(len=FMPI_MAX_PROCESSOR_NAME), intent(out) :: name
    integer, intent(out) :: resultlen
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_get_processor_name
    call mpi_get_processor_name(name, resultlen, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_get_processor_name

  ! fortran bindings for groups, contexts, etc.  ...

  subroutine fmpi_group_size(group, ssize, ierror) ! p. 208
    type(mpi_group_t), intent(in)  :: group
    integer, intent(out) :: ssize
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_size 
    call mpi_group_size (group, ssize, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_size

  subroutine fmpi_group_rank(group, rank, ierror)! p. 208
    type(mpi_group_t), intent(in)  :: group
    integer, intent(out) :: rank
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_rank 
    call mpi_group_rank (group, rank, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_rank

  ! p. 208
  subroutine fmpi_group_translate_ranks(group1,ranks1,group2,ranks2, ierror)
    type(mpi_group_t), intent(in)  :: group1
    integer, intent(in)  :: ranks1(:)
    type(mpi_group_t), intent(in)  :: group2
    integer, intent(out) :: ranks2(size(ranks1))
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_translate_ranks 
    call mpi_group_translate_ranks & 
         (group1, size(ranks1), ranks1, group2, ranks2, ierror_o)    
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_translate_ranks

  subroutine fmpi_group_compare(group1, group2, result, ierror) ! p. 209
    type(mpi_group_t), intent(in)  :: group1
    type(mpi_group_t), intent(in)  :: group2
    integer, intent(out) :: result
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_compare 
    call mpi_group_compare (group1, group2, result, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_compare

  subroutine fmpi_comm_group(group, comm, ierror) ! p. 210
    integer, intent(out) :: group
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_comm_group 
    type(mpi_comm_t) :: comm_o
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_group (comm_o, group, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_group

  subroutine fmpi_group_union(group1, group2, newgroup, ierror) ! p. 211
    type(mpi_group_t), intent(in)  :: group1
    type(mpi_group_t), intent(in)  :: group2
    integer, intent(out) :: newgroup
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_union 
    call mpi_group_union (group1, group2, newgroup, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_union

  ! p. 211
  subroutine fmpi_group_intersection( group1, group2, newgroup, ierror)
    type(mpi_group_t), intent(in)  :: group1
    type(mpi_group_t), intent(in)  :: group2
    integer, intent(out) :: newgroup
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_intersection 
    call mpi_group_intersection (group1, group2, newgroup, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_intersection

  ! p. 211
  subroutine fmpi_group_difference(group1, group2, newgroup, ierror)
    type(mpi_group_t), intent(in)  :: group1
    type(mpi_group_t), intent(in)  :: group2
    integer, intent(out) :: newgroup
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_difference 
    if (present(ierror)) ierror = ierror_o
    call mpi_group_difference ( group1, group2, newgroup, ierror_o)
  end subroutine fmpi_group_difference

  ! p. 212
  subroutine fmpi_group_incl(group, ranks, newgroup, ierror)
    type(mpi_group_t), intent(in)  :: group
    integer, intent(in)  :: ranks(:)
    integer, intent(out) :: newgroup
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_incl 
    call mpi_group_incl (group, size(ranks), ranks, newgroup, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_incl

  ! p. 213
  subroutine fmpi_group_excl(group, ranks, newgroup, ierror)
    type(mpi_group_t), intent(in)  :: group
    integer, intent(in)  :: ranks(:)
    integer, intent(out) :: newgroup
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_excl 
    call mpi_group_excl (group, size(ranks), ranks, newgroup, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_excl

  ! p. 214
  subroutine fmpi_group_range_incl(group, ranges, newgroup, ierror)
    type(mpi_group_t), intent(in)  :: group
    integer, intent(in)  :: ranges(:,:) ! (3,:) but my compiler chokes
    integer, intent(out) :: newgroup
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_range_incl 
    call mpi_group_range_incl &
         ( group, size(ranges,2), ranges, newgroup, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_range_incl

  ! p. 215
  subroutine fmpi_group_range_excl(group, ranges, newgroup, ierror)
    type(mpi_group_t), intent(in)  :: group
    integer, intent(in)  :: ranges(:,:)
    integer, intent(out) :: newgroup
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_range_excl 
    if (present(ierror)) ierror = ierror_o
    call mpi_group_range_excl &
         (group, size(ranges,2), ranges, newgroup, ierror_o)
  end subroutine fmpi_group_range_excl

  ! p. 216
  subroutine fmpi_group_free(group, ierror)
    type(mpi_group_t), intent(inout) :: group
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_group_free 
    call mpi_group_free (group, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_group_free

  ! Section 5.4.1
  subroutine fmpi_comm_size(ssize, comm, ierror)  ! p.217
    integer, intent(out) :: ssize
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_comm_size
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_size(comm_o, ssize, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_size

  subroutine fmpi_comm_rank(rank, comm,  ierror) ! p.217
    integer, intent(out) :: rank
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_comm_rank
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_rank(comm_o, rank, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_rank

  subroutine fmpi_comm_dup(newcomm, comm, ierror)
    type(mpi_comm_t), intent(out) :: newcomm
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o 
    external mpi_comm_dup
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_dup(comm_o, newcomm, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_dup  

  subroutine fmpi_comm_free(comm, ierror)
    type(mpi_comm_t), intent(inout) :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o 
    external mpi_comm_free
    call mpi_comm_free(comm, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_free


  subroutine fmpi_comm_compare(comm1, result,  comm2,  ierror)
    integer, intent(out) :: result
    type(mpi_comm_t), intent(in)  :: comm1
    type(mpi_comm_t), intent(in), optional  :: comm2
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm2_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_comm_compare 
    if (present(comm2)) then
       comm2_o = comm2
    else
       comm2_o = fmpi_comm
    end if
    call mpi_comm_compare (comm1, comm2_o, result, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_compare


  subroutine fmpi_comm_create(group, newcomm, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_comm_t) :: comm_o
    type(mpi_group_t), intent(in)  :: group
    integer, intent(out) :: newcomm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_comm_create 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_create (comm_o, group, newcomm, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_create

  subroutine fmpi_comm_split(color, key, newcomm, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_comm_t) :: comm_o
    integer, intent(in)  :: color
    integer, intent(in)  :: key
    integer, intent(out) :: newcomm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_comm_split 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_split (comm_o, color, key, newcomm, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_split

  subroutine fmpi_comm_test_inter(flag, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t), intent(out), optional :: ierror
    type(mpi_ierr_t) :: ierror_o
    logical :: flag 
    external mpi_comm_test_inter 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_test_inter (comm_o, flag, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_test_inter

  subroutine fmpi_comm_remote_size(ssize, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(out) :: ssize
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    type(mpi_comm_t) :: comm_o
    external mpi_comm_remote_size 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_remote_size (comm_o, ssize, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_remote_size

  subroutine fmpi_comm_remote_group(group, comm, ierror)
    integer, intent(out) :: group
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_comm_remote_group 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_comm_remote_group (comm_o, group, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_comm_remote_group

  subroutine fmpi_intercomm_create(local_comm, local_leader, &
       peer_comm, remote_leader,tag, newintercomm, ierror)
    integer, intent(in)  :: local_comm
    integer, intent(in)  :: local_leader
    integer, intent(in)  :: peer_comm
    integer, intent(in)  :: remote_leader
    integer, intent(in)  :: tag
    integer, intent(out) :: newintercomm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_intercomm_create 
    call mpi_intercomm_create (local_comm, local_leader, &
         peer_comm, remote_leader,tag, newintercomm, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_intercomm_create

  subroutine fmpi_intercomm_merge(intercomm, high, newintracomm, ierror)
    integer, intent(in)  :: intercomm
    logical, intent(in)  :: high 
    integer, intent(out) :: newintracomm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_intercomm_merge 
    call mpi_intercomm_merge (intercomm, high, newintracomm, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_intercomm_merge

  subroutine fmpi_keyval_create(copy_fn, delete_fn, keyval, &
       extra_state, ierror)                                        
    external mpi_keyval_create 
    external copy_fn, delete_fn 
    integer, intent(out) :: keyval
    integer, intent(in)  :: extra_state
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    call mpi_keyval_create(copy_fn, delete_fn, keyval, &
         extra_state, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_keyval_create

  subroutine fmpi_keyval_free(keyval, ierror)
    integer, intent(inout) :: keyval
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_keyval_free 
    call mpi_keyval_free (keyval, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_keyval_free

  subroutine fmpi_attr_put(keyval, attribute_val, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: keyval
    integer, intent(in)  :: attribute_val
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_attr_put 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_attr_put (comm_o, keyval, attribute_val, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_attr_put

  subroutine fmpi_attr_get(keyval, attribute_val, flag, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: keyval
    integer, intent(out) :: attribute_val
    logical, intent(out) :: flag 
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_attr_get 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_attr_get (comm_o, keyval, attribute_val, flag, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_attr_get

  subroutine fmpi_attr_delete(keyval, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: keyval
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_attr_delete 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_attr_delete (comm_o, keyval, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_attr_delete

  !       ... a.12  fortran bindings for process topologies  ...

  subroutine fmpi_cart_create(dims,periods,reorder,comm_cart,comm_old,ierror)
    type(mpi_comm_t), intent(in), optional :: comm_old
    integer, intent(in)  :: dims(:)
    logical, intent(in)  :: periods(size(dims))
    logical, intent(in)  :: reorder 
    type(mpi_comm_t), intent(out) :: comm_cart
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_cart_create 
    if (present(comm_old)) comm_o = comm_old
    call mpi_cart_create (comm_o, size(dims), dims, periods, reorder, &
         comm_cart, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cart_create

  subroutine fmpi_dims_create(nnodes, dims, ierror)
    integer, intent(in)  :: nnodes
    integer, intent(inout) :: dims(:)
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_dims_create 
    call mpi_dims_create (nnodes, size(dims), dims, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_dims_create

  subroutine fmpi_graph_create(index, edges, reorder,comm_graph, &
       comm_old, ierror)    
    integer, intent(in)  :: index(:)
    integer, intent(in)  :: edges(size(index))
    logical, intent(in)  :: reorder 
    type(mpi_comm_t), intent(out) :: comm_graph
    type(mpi_comm_t), intent(in), optional :: comm_old
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_graph_create
    if (present(comm_old)) comm_o = comm_old
    call mpi_graph_create (comm_o, size(index), index, edges, reorder, &
         comm_graph, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_graph_create


!!!!!!!!!!!!!!!!!!!!
  subroutine fmpi_topo_test(status, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(out) :: status
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_topo_test 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_topo_test (status, comm_o, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_topo_test

  subroutine fmpi_graphdims_get(nnodes, nedges, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(out) :: nnodes
    integer, intent(out) :: nedges
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_graphdims_get 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_graphdims_get (comm_o, nnodes, nedges, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_graphdims_get

  subroutine fmpi_graph_get(index, edges, comm, ierror) 
    integer, intent(out) :: index(:)
    integer, intent(out) :: edges(:)
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_graph_get 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_graph_get (comm_o, size(index), size(edges), &
         index, edges, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_graph_get

  subroutine fmpi_cartdim_get(ndims, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(out) :: ndims
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_cartdim_get 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_cartdim_get (comm_o, ndims, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cartdim_get

  subroutine fmpi_cart_get(dims, periods, coords, comm, ierror) 
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(out) :: dims(:)
    logical, intent(out) :: periods(size(dims)) 
    integer, intent(out) :: coords(size(dims))
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_cart_get 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_cart_get (comm_o, size(dims), dims, periods, coords, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cart_get


!!!!!!!!!!!!!!!!!!!!!!
  subroutine fmpi_cart_rank(coords, rank, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: coords(:)
    integer, intent(out) :: rank
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_cart_rank 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_cart_rank (comm_o, coords, rank, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cart_rank

  subroutine fmpi_cart_coords(rank, coords, comm, ierror)
    integer, intent(in)  :: rank
    integer, intent(out) :: coords(:)
    type(mpi_comm_t) :: comm_o
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_cart_coords 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_cart_coords (comm_o, rank, size(coords), coords, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cart_coords

  subroutine fmpi_graph_neighbors_count(rank, nneighbors, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: rank
    integer, intent(out) :: nneighbors
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_graph_neighbors_count 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_graph_neighbors_count (comm_o, rank, nneighbors, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_graph_neighbors_count

  subroutine fmpi_graph_neighbors(rank, neighbors, comm, ierror)           
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: rank
    integer, intent(out) :: neighbors(:)
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_graph_neighbors 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_graph_neighbors (comm_o,rank,size(neighbors), &
         neighbors, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_graph_neighbors


!!!!!!!!!!!!!!!!!!
  subroutine fmpi_cart_shift(direction, disp, rank_source, rank_dest, &
       comm, ierror)                                           
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: direction
    integer, intent(in)  :: disp
    integer, intent(out) :: rank_source
    integer, intent(out) :: rank_dest
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_cart_shift 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_cart_shift ( comm_o, direction, disp, rank_source, &
         rank_dest, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cart_shift

  subroutine fmpi_cart_sub(comm, remain_dims, newcomm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    logical, intent(in)  :: remain_dims(:) 
    integer, intent(out) :: newcomm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_cart_sub 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_cart_sub (comm_o, remain_dims, newcomm, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cart_sub

  subroutine fmpi_cart_map(comm, ndims, dims, periods, newrank, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: ndims
    integer, intent(in)  :: dims(*)
    logical, intent(in)  :: periods(*) 
    integer, intent(out) :: newrank
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_cart_map 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_cart_map ( comm_o, ndims, dims, periods, newrank, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_cart_map

  subroutine fmpi_graph_map( index, edges, newrank, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(in)  :: index(:)
    integer, intent(in)  :: edges(size(index))
    integer, intent(out) :: newrank
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_graph_map 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_graph_map (comm_o,size(index), index, edges, newrank, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_graph_map

!!$
  !       ... a.13  fortran bindings for environmental inquiry  ...


  subroutine fmpi_errhandler_create(function, errhandler, ierror)
    external function 
    type(mpi_errhandler_t), intent(out) :: errhandler
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_errhandler_create 
    call mpi_errhandler_create (function, errhandler, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_errhandler_create

  subroutine fmpi_errhandler_set(errhandler, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_errhandler_t), intent(in) :: errhandler
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_errhandler_set 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_errhandler_set (comm_o, errhandler, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_errhandler_set

  subroutine fmpi_errhandler_get(errhandler, comm, ierror)
    type(mpi_comm_t), intent(in), optional  :: comm
    integer, intent(out) :: errhandler
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o
    external mpi_errhandler_get 
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if
    call mpi_errhandler_get (comm_o, errhandler, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_errhandler_get

  subroutine fmpi_errhandler_free(errhandler, ierror)
    type(mpi_errhandler_t), intent(inout) :: errhandler
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_errhandler_free 
    call mpi_errhandler_free (errhandler, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_errhandler_free

  subroutine fmpi_error_string( errorcode, string, resultlen, ierror)
    integer, intent(in)  :: errorcode
    character(len=*), intent(out) :: string 
    integer, intent(out) :: resultlen
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_error_string 
    call mpi_error_string (errorcode, string, resultlen, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_error_string

  subroutine fmpi_error_class(errorcode, errorclass, ierror)
    integer, intent(in)  :: errorcode
    integer, intent(out) :: errorclass
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_ierr_t) :: ierror_o
    external mpi_error_class 
    call mpi_error_class (errorcode, errorclass, ierror_o) 
    if (present(ierror)) ierror = ierror_o
  end subroutine fmpi_error_class

  !       ... a.14  fortran bindings for profiling  ...

  subroutine fmpi_pcontrol(level)
    integer, intent(in) :: level  !           ... maybe more arguments ...
    external mpi_pcontrol 
    call mpi_pcontrol (level) 
  end subroutine fmpi_pcontrol
end module mpi_main
