! Contains the datatypes used by mpi
! and the overloaded functions for accessing them

module mpi_typedefs
  type mpi_ierr_t
     sequence
     integer :: v
  end type mpi_ierr_t 
  type mpi_comm_t
     integer :: v
  end type mpi_comm_t
  type mpi_group_t
     integer :: v
  end type mpi_group_t
  type mpi_datatype_t
     integer :: v
  end type mpi_datatype_t
  type mpi_request_t
     integer :: v
  end type mpi_request_t
  type mpi_op_t
     integer :: v
  end type mpi_op_t
  type mpi_tag_t
     sequence
     integer :: v
  end type mpi_tag_t
  type mpi_errhandler_t
     integer :: v
  end type mpi_errhandler_t
  type mpi_address_t     ! Addition of addresses is not defined
     integer :: v        ! Subtraction of addresses gives integers
  end type mpi_address_t ! Integers can be added to addresses
                          
  type mpi_status_t
     sequence
     integer         :: count   ! This is known to be correct from mpi.h but is private
     integer         :: source  ! MPI_SOURCE
     type(mpi_tag_t) :: tag     ! MPI_TAG
     type(mpi_ierr_t):: error   ! MPI_ERROR  
  end type mpi_status_t
  public :: calc_displ
  public :: operator(==), operator(/=), operator(+), operator(-) 
  interface operator(==)
     module procedure tag_eq_ss, tag_eq_sa, tag_eq_as, tag_eq_aa, &
          request_eq_ss, request_eq_sa, request_eq_as, request_eq_aa, &
          datatype_eq_ss, datatype_eq_sa, datatype_eq_as, datatype_eq_aa    
  end interface
  interface operator(/=)
     module procedure tag_ne_ss, tag_ne_sa, tag_ne_as, tag_ne_aa, &
          request_ne_ss, request_ne_sa, request_ne_as, request_ne_aa, &
          datatype_ne_ss, datatype_ne_sa, datatype_ne_as, datatype_ne_aa
  end interface
  interface operator(+)
     module procedure adds_plus_is, adda_plus_is, adds_plus_ia, adda_plus_ia, &
          is_plus_adds, ia_plus_adds, is_plus_adda,  ia_plus_adda
  end interface
  interface operator(-)
     module procedure adds_minus_adds,adda_minus_adds,adds_minus_adda,adda_minus_adda
  end interface
contains
  function tag_eq_ss(a,b) result(r)
    type(mpi_tag_t), intent(in) :: a
    type(mpi_tag_t), intent(in) :: b
    logical              :: r
    r = (a%v==b%v)
  end function tag_eq_ss
  function tag_eq_sa(a,b) result(r)
    type(mpi_tag_t), intent(in) :: a
    type(mpi_tag_t), intent(in) :: b(:)
    logical              :: r(size(b))
    r = (a%v==b%v)
  end function tag_eq_sa
  function tag_eq_as(a,b) result(r)
    type(mpi_tag_t), intent(in) :: a(:)
    type(mpi_tag_t), intent(in) :: b
    logical              :: r(size(a))
    r = (a%v==b%v)
  end function tag_eq_as
  function tag_eq_aa(a,b) result(r)
    type(mpi_tag_t), intent(in) :: a(:)
    type(mpi_tag_t), intent(in) :: b(size(a))
    logical              :: r(size(a))
    r = (a%v==b%v)
  end function tag_eq_aa

  function tag_ne_ss(a,b) result(r)
    type(mpi_tag_t), intent(in) :: a
    type(mpi_tag_t), intent(in) :: b
    logical              :: r
    r = (a%v/=b%v)
  end function tag_ne_ss
  function tag_ne_sa(a,b) result(r)
    type(mpi_tag_t), intent(in) :: a
    type(mpi_tag_t), intent(in) :: b(:)
    logical              :: r(size(b))
    r = (a%v/=b%v)
  end function tag_ne_sa
  function tag_ne_as(a,b) result(r)
    type(mpi_tag_t), intent(in) :: a(:)
    type(mpi_tag_t), intent(in) :: b
    logical              :: r(size(a))
    r = (a%v/=b%v)
  end function tag_ne_as
  function tag_ne_aa(a,b) result(r)
    type(mpi_tag_t), intent(in) :: a(:)
    type(mpi_tag_t), intent(in) :: b(size(a))
    logical              :: r(size(a))
    r = (a%v/=b%v)
  end function tag_ne_aa

  function request_eq_ss(a,b) result(r)
    type(mpi_request_t), intent(in) :: a
    type(mpi_request_t), intent(in) :: b
    logical              :: r
    r = (a%v==b%v)
  end function request_eq_ss
  function request_eq_sa(a,b) result(r)
    type(mpi_request_t), intent(in) :: a
    type(mpi_request_t), intent(in) :: b(:)
    logical              :: r(size(b))
    r = (a%v==b%v)
  end function request_eq_sa
  function request_eq_as(a,b) result(r)
    type(mpi_request_t), intent(in) :: a(:)
    type(mpi_request_t), intent(in) :: b
    logical              :: r(size(a))
    r = (a%v==b%v)
  end function request_eq_as
  function request_eq_aa(a,b) result(r)
    type(mpi_request_t), intent(in) :: a(:)
    type(mpi_request_t), intent(in) :: b(size(a))
    logical              :: r(size(a))
    r = (a%v==b%v)
  end function request_eq_aa

  function request_ne_ss(a,b) result(r)
    type(mpi_request_t), intent(in) :: a
    type(mpi_request_t), intent(in) :: b
    logical              :: r
    r = (a%v/=b%v)
  end function request_ne_ss
  function request_ne_sa(a,b) result(r)
    type(mpi_request_t), intent(in) :: a
    type(mpi_request_t), intent(in) :: b(:)
    logical              :: r(size(b))
    r = (a%v/=b%v)
  end function request_ne_sa
  function request_ne_as(a,b) result(r)
    type(mpi_request_t), intent(in) :: a(:)
    type(mpi_request_t), intent(in) :: b
    logical              :: r(size(a))
    r = (a%v/=b%v)
  end function request_ne_as
  function request_ne_aa(a,b) result(r)
    type(mpi_request_t), intent(in) :: a(:)
    type(mpi_request_t), intent(in) :: b(size(a))
    logical              :: r(size(a))
    r = (a%v/=b%v)
  end function request_ne_aa
  ! Subraction of addresses
  ! Addition of integers to addresses
  ! No other operations are defined
  function adds_plus_is(a,b) result (r)
    type(mpi_address_t), intent(in) :: a
    integer, intent(in)           :: b
    type(mpi_address_t)             :: r
    r%v = a%v+b
  end function adds_plus_is
  function adds_plus_ia(a,b) result (r)
    type(mpi_address_t), intent(in) :: a
    integer, intent(in)           :: b(:)
    type(mpi_address_t)             :: r(size(b))
    r%v = a%v+b
  end function adds_plus_ia
  function adda_plus_is(a,b) result (r)
    type(mpi_address_t), intent(in) :: a(:)
    integer, intent(in)           :: b
    type(mpi_address_t)             :: r(size(a))
    r%v = a%v+b
  end function adda_plus_is
  function adda_plus_ia(a,b) result (r)
    type(mpi_address_t), intent(in) :: a(:)
    integer, intent(in)           :: b(size(a))
    type(mpi_address_t)             :: r(size(a))
    r%v = a%v+b
  end function adda_plus_ia

  function is_plus_adds(a,b) result (r)
    integer, intent(in)           :: a
    type(mpi_address_t), intent(in) :: b
    type(mpi_address_t)             :: r
    r%v = a+b%v
  end function is_plus_adds
  function is_plus_adda(a,b) result (r)
    integer, intent(in)           :: a
    type(mpi_address_t), intent(in) :: b(:)
    type(mpi_address_t)             :: r(size(b))
    r%v = a+b%v
  end function is_plus_adda
  function ia_plus_adds(a,b) result (r)
    integer, intent(in)           :: a(:)
    type(mpi_address_t), intent(in) :: b
    type(mpi_address_t)             :: r(size(a))
    r%v = a+b%v
  end function ia_plus_adds
  function ia_plus_adda(a,b) result (r)
    integer, intent(in)           :: a(:)
    type(mpi_address_t), intent(in) :: b(size(a))
    type(mpi_address_t)             :: r(size(a))
    r%v = a+b%v
  end function ia_plus_adda
  
  function adds_minus_adds(a,b) result (r)
   type(mpi_address_t), intent(in) :: a
   type(mpi_address_t), intent(in) :: b
   integer :: r
   r = a%v-b%v
 end function adds_minus_adds
  function adds_minus_adda(a,b) result (r)
   type(mpi_address_t), intent(in) :: a
   type(mpi_address_t), intent(in) :: b(:)
   integer :: r(size(b))
   r = a%v-b%v
 end function adds_minus_adda
  function adda_minus_adds(a,b) result (r)
   type(mpi_address_t), intent(in) :: a(:)
   type(mpi_address_t), intent(in) :: b
   integer :: r(size(a))
   r = a%v-b%v
 end function adda_minus_adds
  function adda_minus_adda(a,b) result (r)
   type(mpi_address_t), intent(in) :: a(:)
   type(mpi_address_t), intent(in) :: b(size(a))
   integer :: r(size(a))
   r = a%v-b%v
 end function adda_minus_adda

  function datatype_eq_ss(a,b) result(r)
    type(mpi_datatype_t), intent(in) :: a
    type(mpi_datatype_t), intent(in) :: b
    logical              :: r
    r = (a%v==b%v)
  end function datatype_eq_ss
  function datatype_eq_sa(a,b) result(r)
    type(mpi_datatype_t), intent(in) :: a
    type(mpi_datatype_t), intent(in) :: b(:)
    logical              :: r(size(b))
    r = (a%v==b%v)
  end function datatype_eq_sa
  function datatype_eq_as(a,b) result(r)
    type(mpi_datatype_t), intent(in) :: a(:)
    type(mpi_datatype_t), intent(in) :: b
    logical              :: r(size(a))
    r = (a%v==b%v)
  end function datatype_eq_as
  function datatype_eq_aa(a,b) result(r)
    type(mpi_datatype_t), intent(in) :: a(:)
    type(mpi_datatype_t), intent(in) :: b(size(a))
    logical              :: r(size(a))
    r = (a%v==b%v)
  end function datatype_eq_aa

  function datatype_ne_ss(a,b) result(r)
    type(mpi_datatype_t), intent(in) :: a
    type(mpi_datatype_t), intent(in) :: b
    logical              :: r
    r = (a%v/=b%v)
  end function datatype_ne_ss
  function datatype_ne_sa(a,b) result(r)
    type(mpi_datatype_t), intent(in) :: a
    type(mpi_datatype_t), intent(in) :: b(:)
    logical              :: r(size(b))
    r = (a%v/=b%v)
  end function datatype_ne_sa
  function datatype_ne_as(a,b) result(r)
    type(mpi_datatype_t), intent(in) :: a(:)
    type(mpi_datatype_t), intent(in) :: b
    logical              :: r(size(a))
    r = (a%v/=b%v)
  end function datatype_ne_as
  function datatype_ne_aa(a,b) result(r)
    type(mpi_datatype_t), intent(in) :: a(:)
    type(mpi_datatype_t), intent(in) :: b(size(a))
    logical              :: r(size(a))
    r = (a%v/=b%v)
  end function datatype_ne_aa
  subroutine calc_displ(d,t,a)
    integer,intent(out) :: d(:)
    integer,intent(out) :: t
    integer,intent(in) :: a(size(d))
    integer i
    d(1)=0
    do i=1,size(d)-1
       d(i+1) = d(i) + a(i)
    end do
    t = d(size(d))+a(size(d))
  end subroutine calc_displ
end module mpi_typedefs
