! This program was inspired by a bug report from 
! fsset@corelli.lerc.nasa.gov (Scott Townsend)
! The original version of this program was submitted by email to 
! mpi-bugs and is in the directory mpich/bugs/ssend (not distributed 
! with the distribution).  This program was modified by William
! Gropp (to correct a few errors and make more consistent with the
! structure of the test programs in the examples/test/pt2pt directory.

! A C version of this program is in allpairc.c
! Modified to use overloaded f90 interface March 1998 Dr. J. McElwaine
!
  !-------------------------------------------------------------------------
  !
  !  Simple pair communication exercises.
  !
  !------------------------------------------------------------------------
program allpair
  use mpi
  implicit none

  integer, parameter :: TEST_SIZE=20
  integer ierror
  type(mpi_tag_t) :: tag
  integer :: prev, next, count, index, i,outcount, indices(2), rank, sz
  type(mpi_status_t) :: status, statuses(2)
  type(mpi_request_t) :: requests(2)
  type(mpi_comm_t) :: dupcom
  logical :: flag
  real, target :: send_buf(TEST_SIZE), recv_buf (TEST_SIZE)
  integer :: dummy(0) ! Zero length array
  call fmpi_init
  call fmpi_comm_rank(rank)
  call fmpi_comm_size(sz)
  call fmpi_comm_dup(dupcom)
  
  if (sz < 2) then
     write(*,'(a)',advance='no') " Allpair needs more than one process."
     call fmpi_abort(41243)
  end if

  next = rank + 1
  if (next >= sz) next = 0
  prev = rank - 1
  if (prev < 0) prev = sz - 1


  !
  !     Normal sends
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Send'
  tag = mpi_tag_t(1123)
  count = TEST_SIZE / 5
  recv_buf=0
  ! Send the messsage from n to n+1, check it, and then repeat  
  ! Send messages of size count
  if (fmpi_is_root(rank)) then  ! start with processor 1
     call init_test_data(send_buf)
     call fmpi_send(send_buf(1:count), next, tag)
     call fmpi_recv(recv_buf,FMPI_ANY_SOURCE, FMPI_ANY_TAG, status)
     call msg_check(recv_buf, prev, tag, count, status, 'send and recv')
  else
     call fmpi_recv(recv_buf,FMPI_ANY_SOURCE, FMPI_ANY_TAG, status)
     call msg_check(recv_buf, prev, tag, count, status, 'send and recv')
     call fmpi_Send(recv_buf(1:count), next, tag)
  end if

  !
  !     Ready sends.  Note that we must insure that the receive is posted
  !     before the rsend; this requires using Irecv.
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Rsend'
  tag = mpi_tag_t(1456)
  count = TEST_SIZE / 3
  recv_buf=0
!  call mpi_irecv(recv_buf, size(recv_buf), MPI_REAL, MPI_ANY_SOURCE, MPI_ANY_TAG, &
!       MPI_COMM_WORLD, requests(1), ierror)
  call fmpi_irecv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG,  requests(1))
!  call test_irecv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG, requests(1))

  if (fmpi_is_root(rank)) then
     call init_test_data(send_buf)
     call fmpi_rsend(send_buf(1:count), next, tag) 
     do 
        call fmpi_test(requests(1), flag, status)
        if (flag) exit
     end do
 !  call fmpi_Wait(requests(1), status)
    call msg_check(recv_buf, prev, tag, count, status, 'rsend and recv')
  else
     do 
        call fmpi_test(requests(1), flag, status)
        if (flag) exit
     end do
 !  call fmpi_Wait(requests(1), status)
     call msg_check(recv_buf, prev, tag, count, status, 'rsend and recv' )
     call fmpi_Send(recv_buf(1:count), next, tag) 
  end if


  !     Synchronous sends
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Ssend'
  tag = mpi_tag_t(1789)
  recv_buf=0
  if (fmpi_is_root(rank)) then
     call init_test_data(send_buf)
     call fmpi_Iprobe(FMPI_ANY_SOURCE, tag, flag, status) 
     if (flag) then
        write(*,'(a)',advance='no') 'Iprobe succeeded! source', status%source, ', tag', status%tag
     end if
     call fmpi_Ssend(send_buf(1:count), next, tag) 
     do 
        if (flag) exit
        call fmpi_Iprobe(FMPI_ANY_SOURCE, tag, flag, status) 
     end do
     if (status%source /= prev) then
        write(*,'(a)',advance='no') 'Incorrect source, expected', prev,', got', status%source
     end if
     if (status%tag /= tag) then
        write(*,'(a)',advance='no') 'Incorrect tag, expected', tag,', got', status%tag
     end if
     call fmpi_get_count(status, send_buf, i)
     if (i /= count) then
        write(*,'(a)',advance='no') 'Incorrect count, expected', count, ', got', i
     end if
     call fmpi_Recv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG, status)
     call msg_check(recv_buf, prev, tag, count, status, 'ssend and recv' ) 
  else
     call fmpi_Recv(recv_buf,FMPI_ANY_SOURCE, FMPI_ANY_TAG,status)
     call msg_check( recv_buf, prev, tag, count, status, 'ssend and recv' )
     call fmpi_Ssend(recv_buf(1:count), next, tag) 
  end if

  !
  !     Nonblocking normal sends
  !
  if (fmpi_is_root(rank))  write(*,'(a)',advance='no')'    Isend'
  tag = mpi_tag_t(2123)
  count = TEST_SIZE / 5
  recv_buf=0
  if (fmpi_is_root(rank)) then
     call fmpi_irecv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG,  requests(1))
     call init_test_data(send_buf)
     call fmpi_Isend(send_buf(1:count), next, tag, requests(2)) 
     call fmpi_Wait(requests(1), statuses(1))
     call fmpi_Wait(requests(2), statuses(2))
     call rq_check(requests, 'isend and irecv' )
     call msg_check(recv_buf, prev, tag, count, statuses(1),'isend and irecv')
  else
     call fmpi_Recv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG,status)
     call msg_check(recv_buf, prev, tag, count, status,'isend and irecv' )
     call fmpi_Isend(recv_buf(1:count), next, tag, requests(1)) 
     call fmpi_Wait(requests(1), status)
     call rq_check(requests(1:1), 'isend and irecv' )
  end if
  !
  !     Nonblocking ready sends
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Irsend'
  tag = mpi_tag_t(2456)
  count = TEST_SIZE / 3
  recv_buf=0
  call fmpi_Irecv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG, requests(1))
  call fmpi_barrier
  call fmpi_Sendrecv(dummy,next,mpi_tag_t(0),dummy,prev,mpi_tag_t(0), comm=dupcom)
  if (fmpi_is_root(rank)) then
     call init_test_data(send_buf)
     call fmpi_Irsend(send_buf(1:count), next, tag, requests(2)) 
     call fmpi_wait(requests(1), status)
     call rq_check(requests(1:1), 'irsend and irecv')
     call msg_check(recv_buf, prev, tag, count, status, 'irsend and irecv')
  else
     call fmpi_wait(requests(1), status)
     call rq_check( requests(1:1), 'irsend and irecv (test)' )
     call msg_check(recv_buf, prev, tag, count, status,'irsend and irecv' )
     call fmpi_irsend(recv_buf(1:count), next, tag, requests(1)) 
     call fmpi_wait(requests(1))
     call rq_check( requests(1:1), 'irsend and irecv' )
  end if
  !
  !     Nonblocking synchronous sends
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Issend'

  tag = mpi_tag_t(2789)
  count = TEST_SIZE / 3
  recv_buf=0
  if (fmpi_is_root(rank)) then
     call fmpi_Irecv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG, requests(1))
     call init_test_data(send_buf)
     call fmpi_Issend(send_buf(1:count), next, tag, requests(2)) 
     do 
        call fmpi_Testall(requests, flag, statuses)
        if (flag) exit
     end do
     call rq_check( requests, 'issend and irecv (testall)' )
     call msg_check( recv_buf, prev, tag, count, statuses(1), 'issend and recv' )
  else
     call fmpi_Recv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG, status)
     call msg_check(recv_buf, prev, tag, count, status, 'issend and recv')
     call fmpi_Issend(recv_buf(1:count), next, tag, requests(1)) 
     do 
        call fmpi_Testany(requests(1:1), index, flag, statuses(1))
        if (flag) exit
     end do
     call rq_check(requests(1:1), 'issend and recv (testany)')
  end if
  !
  !     Persistent normal sends
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Send_init' 
  tag = mpi_tag_t(3123)
  count = TEST_SIZE / 5
  recv_buf=0
!  call fmpi_Send_init(send_buf(1:count), next, tag,requests(1)) 
!  call fmpi_Recv_init(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG, requests(2))
  call mpi_Send_init(send_buf, count, MPI_REAL, next, tag, &
       MPI_COMM_WORLD, requests(1), ierror) 
  call mpi_Recv_init(recv_buf, size(recv_buf), MPI_REAL, MPI_ANY_SOURCE, &
       MPI_ANY_TAG, MPI_COMM_WORLD,  requests(2), ierror)
  if (fmpi_is_root(rank)) then
     call init_test_data(send_buf)
     call fmpi_Startall(requests) 
     call fmpi_Wait(requests(1), statuses(1))
     call fmpi_Wait(requests(2), statuses(2))
!     call fmpi_Waitall(requests, statuses)
!call mpi_wait(requests(1), statuses(1), ierror)
!call mpi_wait(requests(2), statuses(2), ierror)
!call mpi_waitall(size(requests),requests, statuses, ierror)
     call msg_check(recv_buf, prev, tag, count, statuses(2), 'persistent send/recv' )
  else
     call fmpi_Start(requests(2)) 
     call fmpi_Wait(requests(2), status)
     call msg_check( recv_buf, prev, tag, count, status, 'persistent send/recv')
     send_buf = recv_buf
     call fmpi_Start(requests(1)) 
     call fmpi_Wait(requests(1), status)
  end if
  call fmpi_Request_free(requests(1))
  call fmpi_Request_free(requests(2))
  !
  !     Persistent ready sends
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Rsend_init'
  tag = mpi_tag_t(3456)
  count = TEST_SIZE / 3
  recv_buf=0
  call mpi_rSend_init(send_buf, count, MPI_REAL, next, tag, &
       MPI_COMM_WORLD, requests(1), ierror) 
  call mpi_Recv_init(recv_buf, size(recv_buf), MPI_REAL, MPI_ANY_SOURCE, &
       MPI_ANY_TAG, MPI_COMM_WORLD,  requests(2), ierror)
!  call fmpi_Rsend_init(send_buf(1:count), next, tag,requests(1)) 
!  call fmpi_Recv_init(recv_buf,FMPI_ANY_SOURCE, FMPI_ANY_TAG,requests(2))
  if (fmpi_is_root(rank)) then
     call init_test_data(send_buf(1:count))
     call fmpi_startall(requests)
     do 
        call fmpi_waitsome(requests, outcount, indices, statuses)
        if (any(indices(1:outcount)==2)) exit
     end do
     if (indices(1)==2) then
        call msg_check(recv_buf, prev, tag, count, statuses(1), 'waitsome')
     else
        call msg_check(recv_buf, prev, tag, count, statuses(2), 'waitsome')
     end if
  else
     call fmpi_Start(requests(2))
     call fmpi_Wait(requests(2), status)
!     do 
!        call fmpi_Test(requests(2), flag, status)
!        if (flag) exit
!     end do
     call msg_check(recv_buf, prev, tag, count, status,'test')
     send_buf(1:count) = recv_buf(1:count)
     call fmpi_Start(requests(1))
     call fmpi_Wait(requests(1))
  end if
  call fmpi_Request_free(requests(1))
  call fmpi_Request_free(requests(2))
  !
  !     Persistent synchronous sends
  !
  if (fmpi_is_root(rank))write(*,'(a)',advance='no') '    Ssend_init'
  tag = mpi_tag_t(3789)
  count = TEST_SIZE / 3
  recv_buf=0
!  call fmpi_Ssend_init(send_buf(1:count), next, tag, requests(2)) 
!  call fmpi_Recv_init(recv_buf,FMPI_ANY_SOURCE, FMPI_ANY_TAG, requests(1))
  call mpi_sSend_init(send_buf, count, MPI_REAL, next, tag, &
       MPI_COMM_WORLD, requests(2), ierror) 
  call mpi_Recv_init(recv_buf, size(recv_buf), MPI_REAL, MPI_ANY_SOURCE, &
       MPI_ANY_TAG, MPI_COMM_WORLD,  requests(1), ierror)
  if (fmpi_is_root(rank)) then
     call init_test_data(send_buf)
     call fmpi_Startall(requests)
     do 
        call fmpi_Testsome(requests, outcount, indices, statuses)
        if (any(indices(1:outcount) == 1)) exit
     end do
     if (indices(1)==1) then
        call msg_check(recv_buf, prev, tag, count, statuses(1), 'waitsome')
     else
        call msg_check(recv_buf, prev, tag, count, statuses(2), 'waitsome')
     end if
  else
     call fmpi_Start(requests(1))
     do 
        call fmpi_Testany(requests(1:1), index, flag, statuses(1))
        if (flag) exit
     end do
     call msg_check(recv_buf, prev, tag, count, statuses(1), 'testany' )
     send_buf(1:count) = recv_buf(1:count)
     call fmpi_Start(requests(2))
     call fmpi_Wait(requests(2))
  end if
  call fmpi_Request_free(requests(1))
  call fmpi_Request_free(requests(2))
  !
  !     Send/receive.
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Sendrecv'
  tag = mpi_tag_t(4123)
  count = TEST_SIZE / 5
  recv_buf=0
  if (fmpi_is_root(rank)) then
     call init_test_data(send_buf)
     call fmpi_Sendrecv(send_buf(1:count),next,tag,&
          recv_buf(1:count),prev,tag,status) 
     call msg_check(recv_buf, prev, tag, count, status, 'sendrecv' )
  else
     call fmpi_Recv(recv_buf, FMPI_ANY_SOURCE, FMPI_ANY_TAG, status)
     call msg_check(recv_buf, prev, tag, count, status, 'recv/send' )
     call fmpi_Send(recv_buf(1:count), next, tag) 
  end if
  !
  !     Send/receive replace.
  !
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no') '    Sendrecv_replace'
  tag = mpi_tag_t(4456)
  count = TEST_SIZE / 3
  if (fmpi_is_root(rank)) then
     call init_test_data(recv_buf(1:count))
     recv_buf(count+1:) = 0
     call fmpi_Sendrecv_replace(recv_buf(1:count), next, tag, prev, tag, status)  
     call msg_check( recv_buf, prev, tag, count, status,'sendrecvreplace' )
  else
     recv_buf=0
     call fmpi_Recv(recv_buf,FMPI_ANY_SOURCE, FMPI_ANY_TAG,status)
     call msg_check( recv_buf, prev, tag, count, status,'recv/send for replace' )
     call fmpi_Send(recv_buf(1:count), next, tag) 
  end if
  call fMPI_Comm_free(dupcom)
  if (fmpi_is_root(rank)) then
     write(*,*)
     write(*,'(a)')" allpair test successful"
  end if
  call fmpi_finalize
contains 
  !------------------------------------------------------------------------------
  !
  !  Check for correct source, tag, count, and data in test message.
  ! The source, tag and count are passed in and checked against the status
  !
  !------------------------------------------------------------------------------
  subroutine msg_check(recv_buf, source, tag, count, status, name)
    real, intent(in)               :: recv_buf(:)
    integer, intent(in)            :: source
    type(mpi_tag_t), intent(in)    :: tag
    integer, intent(in)            :: count
    type(mpi_status_t), intent(in) :: status
    character(len=*), intent(in)   :: name
    integer rank, recv_count
    call fMPI_Comm_rank(rank)
    call fMPI_Get_count(status, recv_buf, recv_count)
    if (status%source /= source) then
       write(*,*)
       write(*,*) '[', rank, '] Unexpected source:', status%source, ' in ', name
       write(*,*)'source, tag, count',source, tag, count
       write(*,*)'status',status
!       call fMPI_Abort(101)
    end if
    if (status%tag /= tag) then
       write(*,*)
       write(*,*) '[', rank, '] Unexpected tag:', status%tag, ' in ', name
       write(*,*)'source, tag, count',source, tag, count
       write(*,*)'status',status
!       call fMPI_Abort(102)
    end if
    if (recv_count /= count) then
       write(*,*)
       write(*,*) '[', rank, '] Unexpected count:', recv_count,' in ', name
       write(*,*)'source, tag, count',source, tag, count
       write(*,*)'status',status
!       call fMPI_Abort(103)
    end if
    call verify_test_data(recv_buf, count, name, status)
  end subroutine msg_check
  !------------------------------------------------------------------------------
  !
  !  Check that requests have been set to null
  !
  !------------------------------------------------------------------------------
  subroutine rq_check(requests, msg )
    type(mpi_request_t), intent(in) :: requests(:)
    character(len=*), intent(in) :: msg
    if (any(requests /= FMPI_REQUEST_NULL)) write(*,*) 'Nonnull request in ', msg
  end subroutine rq_check
  !------------------------------------------------------------------------------
  !
  !  Initialize test data buffer with integral sequence.
  !
  !------------------------------------------------------------------------------
  subroutine init_test_data(buf)
    real, intent(out) :: buf(:)
    integer i
    do  i = 1, size(buf)
       buf(i) = real(i)
    end do
  end subroutine init_test_data
  !------------------------------------------------------------------------------
  !
  !  Verify test data buffer
  !
  !------------------------------------------------------------------------------
  subroutine verify_test_data(buf, count, name, status)
    real, intent(in) :: buf(:)
    character (len=*), intent(in) :: name
    integer, intent(in) :: count
    type(mpi_status_t), intent(in), optional :: status
    integer :: i
    logical found_err
    found_err = .false.
    do  i = 1, count
       if (buf(i) /= REAL(i)) then  
           write(*,*) 'Invalid data value', buf(i), 'at', i, 'of', count ,'in ',name
           found_err = .true.
       endif
    end do
    if (any(buf(count + 1:)/=0)) then
       found_err = .true.
       write(*,*) 'End of the receive buffer has been overwritten', name
    endif
    if (found_err) then
       if (present(status)) then
          write(*,*)'status%count',status%count
          write(*,*)'status%source',status%source
          write(*,*)'status%tag',status%tag
          write(*,*)'status%error',status%error
       end if
       write(*,*)'Error is on node', fmpi_rank
       !call fMPI_Abort(109)
     end if
  end subroutine verify_test_data
  subroutine test_irecv(buf, source, tag, request, comm, ierror)
    real, target, intent(out) :: buf(:)
    integer, intent(in)  :: source
    type(mpi_tag_t), intent(in)  :: tag
    type(mpi_request_t), intent(out) :: request
    type(mpi_comm_t), intent(in), optional  :: comm
    type(mpi_ierr_t), intent(out), optional :: ierror 
    type(mpi_comm_t) :: comm_o
    type(mpi_ierr_t) :: ierror_o 
    external mpi_irecv
    if (present(comm)) then
       comm_o = comm    
    else
       comm_o = fmpi_comm
    end if

    call mpi_irecv(buf(1), size(buf), fmpi_real, source, tag, comm_o, request, ierror_o)
    if (present(ierror)) ierror = ierror_o
  end subroutine test_irecv
end program allpair
