!
!
program main
  use mpi
  implicit none

  integer count, errcnt, size, rank, i
  integer comm
  logical fnderr
  integer max_size
  integer world_rank
  parameter (max_size=100)
  integer intin(max_size), intout(max_size), intsol(max_size)
  real    realin(max_size), realout(max_size), realsol(max_size)
  double precision dblein(max_size), dbleout(max_size),dblesol(max_size)
  complex cplxin(max_size), cplxout(max_size), cplxsol(max_size)
  logical login(max_size), logout(max_size), logsol(max_size)

  !
  !       declare work areas
  !
  call fmpi_init

  errcnt = 0
  call fmpi_comm_rank(rank)
  world_rank = rank
  call fmpi_comm_size(sz)

  ! test sum 
  if (fmpi_is_root(rank)) print *, ' mpi_sum'

  fnderr = .false.

  intin = (/i,i=1,size(intin))
  intsol = sz * intin
  intout = 0

  call fmpi_allreduce(intin, intout, mpi_sum)
  if (any(intout(i)/=intsol)) then
     errcnt = errcnt + 1
     print *, 'error for type mpi_integer and op mpi_sum'
  endif

  intin = (/i,i=1,size(intin))
  intsol = sz * intin
  intout = 0

  call fmpi_allreduce(realin, realout, mpi_sum)
  if (any(realout/=realsol)) then
        errcnt = errcnt + 1
     print *, 'error for type mpi_real and op mpi_sum'
  endif

  dblein = (/i,i=1,size(intin))
  dblesol = sz * intin
  dbleout = 0
  call fmpi_allreduce(dblein, dbleout,  mpi_sum)
  if (any(dbleout/=dblesol)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     print *, 'error for type mpi_double_precision and op mpi_sum'
  endif


  cplxin = (/i,i=1,size(intin))
  cplxsol = sz * intin
  cplxout = 0
  call fmpi_allreduce( cplxin, cplxout,  mpi_sum)
  if (any(cplxout /= cplxsol) then
        errcnt = errcnt + 1
     print *, 'error for type mpi_complex and op mpi_sum'
  endif


  if (errcnt > 0) then
     print *, 'found ', errcnt, ' errors on ', rank, ' for mpi_sum'
     call fmpi_abort(701)
  endif
  errcnt = 0

  ! test product 
  if (fmpi_is_root(rank)) print *, ' mpi_prod'

  fnderr = .false.
  do  i=1,count
     intin(i) = i
     intsol(i) = (i)**(sz)
     intout(i) = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_integer, mpi_prod)
  do  i=1,count
     if (intout(i).ne.intsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_integer and op mpi_prod'
  endif


  fnderr = .false.
  do  i=1,count
     realin(i) = i
     realsol(i) = (i)**(sz)
     realout(i) = 0
  end do
  call fmpi_allreduce( realin, realout, count, 
  *      mpi_real, mpi_prod)
  do  i=1,count
     if (realout(i).ne.realsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_real and op mpi_prod'
  endif


  fnderr = .false.
  do  i=1,count
     dblein(i) = i
     dblesol(i) = (i)**(sz)
     dbleout(i) = 0
  end do
  call fmpi_allreduce( dblein, dbleout, count, 
  *      mpi_double_precision, mpi_prod)
  do  i=1,count
     if (dbleout(i).ne.dblesol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_double_precision and op mpi_prod'
  endif


  fnderr = .false.
  do  i=1,count
     cplxin(i) = i
     cplxsol(i) = (i)**(sz)
     cplxout(i) = 0
  end do
  call fmpi_allreduce( cplxin, cplxout, count, 
  *      mpi_complex, mpi_prod)
  do  i=1,count
     if (cplxout(i).ne.cplxsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_complex and op mpi_prod'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, ' for mpi_prod'
  endif
  errcnt = 0

  !  test max
  if (fmpi_is_root(rank)) print *, ' mpi_max'

  fnderr = .false.
  do  i=1,count
     intin(i) = (rank + i)
     intsol(i) = (sz - 1 + i)
     intout(i) = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_integer, mpi_max)
  do  i=1,count
     if (intout(i).ne.intsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_integer and op mpi_max'
  endif


  fnderr = .false.
  do  i=1,count
     realin(i) = (rank + i)
     realsol(i) = (sz - 1 + i)
     realout(i) = 0
  end do
  call fmpi_allreduce( realin, realout, count, 
  *      mpi_real, mpi_max)
  do  i=1,count
     if (realout(i).ne.realsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_real and op mpi_max'
  endif


  fnderr = .false.
  do  i=1,count
     dblein(i) = (rank + i)
     dblesol(i) = (sz - 1 + i)
     dbleout(i) = 0
  end do
  call fmpi_allreduce( dblein, dbleout, count, 
  *      mpi_double_precision, mpi_max)
  do  i=1,count
     if (dbleout(i).ne.dblesol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_double_precision and op mpi_max'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, ' for mpi_max'
  endif
  errcnt = 0

  ! test min 
  if (fmpi_is_root(rank)) print *, ' mpi_min'

  fnderr = .false.
  do  i=1,count
     intin(i) = (rank + i)
     intsol(i) = i
     intout(i) = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_integer, mpi_min)
  do  i=1,count
     if (intout(i).ne.intsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_integer and op mpi_min'
  endif


  fnderr = .false.
  do  i=1,count
     realin(i) = (rank + i)
     realsol(i) = i
     realout(i) = 0
  end do
  call fmpi_allreduce( realin, realout, count, 
  *      mpi_real, mpi_min)
  do  i=1,count
     if (realout(i).ne.realsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_real and op mpi_min'
  endif


  fnderr = .false.
  do  i=1,count
     dblein(i) = (rank + i)
     dblesol(i) = i
     dbleout(i) = 0
  end do
  call fmpi_allreduce( dblein, dbleout, count, 
  *      mpi_double_precision, mpi_min)
  do  i=1,count
     if (dbleout(i).ne.dblesol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_double_precision and op mpi_min'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, ' for mpi_min'
  endif
  errcnt = 0

  ! test lor
  if (fmpi_is_root(rank)) print *, ' mpi_lor'

  fnderr = .false.
  do  i=1,count
     login(i) = (mod(rank,2) .eq. 1)
     logsol(i) = (sz .gt. 1)
     logout(i) = .false.
  end do
  call fmpi_allreduce( login, logout, count, 
  *      mpi_logical, mpi_lor)
  do  i=1,count
     if (logout(i).neqv.logsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_logical and op mpi_lor'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, ' for mpi_lor(0)'
  endif
  errcnt = 0



  fnderr = .false.
  do  i=1,count
     login(i) = .false.
     logsol(i) = .false.
     logout(i) = .false.
  end do
  call fmpi_allreduce( login, logout, count, 
  *      mpi_logical, mpi_lor)
  do  i=1,count
     if (logout(i).neqv.logsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_logical and op mpi_lor'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, ' for mpi_lor(1)'
  endif
  errcnt = 0

  ! test lxor 
  if (fmpi_is_root(rank)) print *, ' mpi_lxor'

  fnderr = .false.
  do  i=1,count
     login(i) = (rank .eq. 1)
     logsol(i) = (sz .gt. 1)
     logout(i) = .false.
  end do
  call fmpi_allreduce( login, logout, count, 
  *      mpi_logical, mpi_lxor)
  do  i=1,count
     if (logout(i).neqv.logsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_logical and op mpi_lxor'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ',errcnt,' errors on ', rank, ' for mpi_lxor'
  endif
  errcnt = 0


  fnderr = .false.
  do  i=1,count
     login(i) = .false.
     logsol(i) = .false.
     logout(i) = .false.
  end do
  call fmpi_allreduce( login, logout, count, 
  *      mpi_logical, mpi_lxor)
  do  i=1,count
     if (logout(i).neqv.logsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_logical and op mpi_lxor'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ',errcnt,' errors on ',rank,' for mpi_lxor(0)'
  endif
  errcnt = 0


  fnderr = .false.
  do  i=1,count
     login(i) = .true.
     logsol(i) = .false.
     logout(i) = .false.
  end do
  call fmpi_allreduce( login, logout, count, 
  *      mpi_logical, mpi_lxor)
  do  i=1,count
     if (logout(i).neqv.logsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_logical and op mpi_lxor'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ',errcnt,' errors on ',rank,' for mpi_lxor(1-0)'
  endif
  errcnt = 0

  ! test land 
  if (fmpi_is_root(rank)) print *, ' mpi_land'

  fnderr = .false.
  do  i=1,count
     login(i) = (mod(rank,2) .eq. 1)
     logsol(i) = .false.
     logout(i) = .false.
  end do
  call fmpi_allreduce( login, logout, count, 
  *      mpi_logical, mpi_land)
  do  i=1,count
     if (logout(i).neqv.logsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_logical and op mpi_land'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, ' for mpi_land'
  endif
  errcnt = 0




  fnderr = .false.
  do  i=1,count
     login(i) = .true.
     logsol(i) = .true.
     logout(i) = .false.
  end do
  call fmpi_allreduce( login, logout, count, 
  *      mpi_logical, mpi_land)
  do  i=1,count
     if (logout(i).neqv.logsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_logical and op mpi_land'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ',errcnt,' errors on ',rank,
     *      ' for mpi_land(true)'
  endif
  errcnt = 0

  ! test bor
  if (fmpi_is_root(rank)) print *, ' mpi_bor'
  if (sz .lt. 3) then

     fnderr = .false.
     do  i=1,count
        intin(i) = mod(rank,4)
        intsol(i) = sz - 1
        intout(i) = 0
     end do
     call fmpi_allreduce( intin, intout, count, 
     *      mpi_integer, mpi_bor)
     do  i=1,count
        if (intout(i).ne.intsol(i)) then
           errcnt = errcnt + 1
           fnderr = .true. 
        endif
     end do
     if (fnderr) then
        print *, 'error for type mpi_integer and op mpi_bor'
     endif

  else

     fnderr = .false.
     do  i=1,count
        intin(i) = mod(rank,4)
        intsol(i) = 3
        intout(i) = 0
     end do
     call fmpi_allreduce( intin, intout, count, 
     *      mpi_integer, mpi_bor)
     do  i=1,count
        if (intout(i).ne.intsol(i)) then
           errcnt = errcnt + 1
           fnderr = .true. 
        endif
     end do
     if (fnderr) then
        print *, 'error for type mpi_integer and op mpi_bor'
     endif

  endif
  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank,
     *             ' for mpi_bor(1)'
  endif
  errcnt = 0

  ! test band 
  if (fmpi_is_root(rank)) print *, ' mpi_band'
  ! see bottom for function definitions

  fnderr = .false.
  do  i=1,count
     intin(i) = ibxandval(rank,sz,i)
     intsol(i) = i
     intout(i) = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_integer, mpi_band)
  do  i=1,count
     if (intout(i).ne.intsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_integer and op mpi_band'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, 
     *            ' for mpi_band(1)'
  endif
  errcnt = 0


  fnderr = .false.
  do  i=1,count
     intin(i) = ibxandval1(rank,sz,i)
     intsol(i) = 0
     intout(i) = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_integer, mpi_band)
  do  i=1,count
     if (intout(i).ne.intsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_integer and op mpi_band'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, 
     *            ' for mpi_band(0)'
  endif
  errcnt = 0

  ! test bxor 
  if (fmpi_is_root(rank)) print *, ' mpi_bxor'
  ! see below for function definitions

  fnderr = .false.
  do  i=1,count
     intin(i) = ibxorval1(rank)
     intsol(i) = ibxorsol1(sz)
     intout(i) = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_integer, mpi_bxor)
  do  i=1,count
     if (intout(i).ne.intsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_integer and op mpi_bxor'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, 
     *            ' for mpi_bxor(1)'
  endif
  errcnt = 0


  fnderr = .false.
  do  i=1,count
     intin(i) = 0
     intsol(i) = 0
     intout(i) = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_integer, mpi_bxor)
  do  i=1,count
     if (intout(i).ne.intsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_integer and op mpi_bxor'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, 
     *            ' for mpi_bxor(0)'
  endif
  errcnt = 0

  ! assumes -1 == all bits set

  fnderr = .false.
  do  i=1,count
     intin(i) = (-1)
     intsol(i) = 0
     intout(i) = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_integer, mpi_bxor)
  do  i=1,count
     if (intout(i).ne.intsol(i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_integer and op mpi_bxor'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, 
     *            ' for mpi_bxor(1-0)'
  endif
  errcnt = 0

  ! test maxloc 
  if (fmpi_is_root(rank)) print *, ' mpi_maxloc'

  fnderr = .false.
  do  i=1, count
     intin(2*i-1) = (rank + i)
     intin(2*i)   = rank
     intsol(2*i-1) = (sz - 1 + i)
     intsol(2*i) = (sz-1)
     intout(2*i-1) = 0
     intout(2*i)   = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_2integer, mpi_maxloc)
  do  i=1, count
     if (intout(2*i-1) .ne. intsol(2*i-1) .or.
        *      intout(2*i) .ne. intsol(2*i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_2integer and op mpi_maxloc'
  endif


  fnderr = .false.
  do  i=1, count
     realin(2*i-1) = (rank + i)
     realin(2*i)   = rank
     realsol(2*i-1) = (sz - 1 + i)
     realsol(2*i) = (sz-1)
     realout(2*i-1) = 0
     realout(2*i)   = 0
  end do
  call fmpi_allreduce( realin, realout, count, 
  *      mpi_2real, mpi_maxloc)
  do  i=1, count
     if (realout(2*i-1) .ne. realsol(2*i-1) .or.
        *      realout(2*i) .ne. realsol(2*i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_2real and op mpi_maxloc'
  endif


  fnderr = .false.
  do  i=1, count
     dblein(2*i-1) = (rank + i)
     dblein(2*i)   = rank
     dblesol(2*i-1) = (sz - 1 + i)
     dblesol(2*i) = (sz-1)
     dbleout(2*i-1) = 0
     dbleout(2*i)   = 0
  end do
  call fmpi_allreduce( dblein, dbleout, count, 
  *      mpi_2double_precision, mpi_maxloc)
  do  i=1, count
     if (dbleout(2*i-1) .ne. dblesol(2*i-1) .or.
        *      dbleout(2*i) .ne. dblesol(2*i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_2double_precision and op mpi_maxloc'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, 
     *            ' for mpi_maxloc'
  endif
  errcnt = 0

  ! test minloc 
  if (fmpi_is_root(rank)) print *, ' mpi_minloc'


  fnderr = .false.
  do  i=1, count
     intin(2*i-1) = (rank + i)
     intin(2*i)   = rank
     intsol(2*i-1) = i
     intsol(2*i) = 0
     intout(2*i-1) = 0
     intout(2*i)   = 0
  end do
  call fmpi_allreduce( intin, intout, count, 
  *      mpi_2integer, mpi_minloc)
  do  i=1, count
     if (intout(2*i-1) .ne. intsol(2*i-1) .or.
        *      intout(2*i) .ne. intsol(2*i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_2integer and op mpi_minloc'
  endif


  fnderr = .false.
  do  i=1, count
     realin(2*i-1) = (rank + i)
     realin(2*i)   = rank
     realsol(2*i-1) = i
     realsol(2*i) = 0
     realout(2*i-1) = 0
     realout(2*i)   = 0
  end do
  call fmpi_allreduce( realin, realout, count, 
  *      mpi_2real, mpi_minloc)
  do  i=1, count
     if (realout(2*i-1) .ne. realsol(2*i-1) .or.
        *      realout(2*i) .ne. realsol(2*i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_2real and op mpi_minloc'
  endif


  fnderr = .false.
  do  i=1, count
     dblein(2*i-1) = (rank + i)
     dblein(2*i)   = rank
     dblesol(2*i-1) = i
     dblesol(2*i) = 0
     dbleout(2*i-1) = 0
     dbleout(2*i)   = 0
  end do
  call fmpi_allreduce( dblein, dbleout, count, 
  *      mpi_2double_precision, mpi_minloc)
  do  i=1, count
     if (dbleout(2*i-1) .ne. dblesol(2*i-1) .or.
        *      dbleout(2*i) .ne. dblesol(2*i)) then
        errcnt = errcnt + 1
        fnderr = .true. 
     endif
  end do
  if (fnderr) then
     print *, 'error for type mpi_2double_precision and op mpi_minloc'
  endif


  if (errcnt .gt. 0) then
     print *, 'found ', errcnt, ' errors on ', rank, 
     *            ' for mpi_minloc'
  endif
  errcnt = 0

  call fmpi_finalize

contains
  integer function ibxorval1( ir )
  ibxorval1 = 0
  if (ir .eq. 1) ibxorval1 = 16+32+64+128
  return
end program main

integer function ibxorsol1( is )
ibxorsol1 = 0
if (is .gt. 1) ibxorsol1 = 16+32+64+128
return
      end

!
!      assumes -1 == all bits set
      integer function ibxandval( ir, is, i )
      integer ir, is, i
      ibxandval = -1
      if (ir .eq. is - 1) ibxandval = i
      return
      end
!
      integer function ibxandval1( ir, is, i )
      integer ir, is, i
      ibxandval1 = 0
      if (ir .eq. is - 1) ibxandval1 = i
      return
      end
 
end program allredf
