! Adapted from the allredf program in the mpich distribution
program allreduce
  use mpi
  implicit none
  integer, parameter :: count = 10
  integer rank

  call fmpi_init
  call fmpi_comm_rank(rank)
  if (fmpi_is_root(rank)) then
     write(*,'(a)',advance='no')' Testing integers: '
  end if
  call test_s_integer_reductions   
  call test_integer_reductions(count)
  call test_integer_only_reductions(count)
  if (fmpi_is_root(rank)) then
     write(*,*)
     write(*,'(a)',advance='no')' Testing reals:    '
  end if
  call test_real_reductions(count)
  if (fmpi_is_root(rank))  then
     write(*,*)
     write(*,'(a)',advance='no')' Testing logicals: '
  end if
  call test_logical_reductions(count)
  call fmpi_finalize
contains
  subroutine check_ans_s(ans,msg)
    logical :: ans
    character(len=*) :: msg
    if (.not.ans) then
       print *, msg
       call fmpi_abort(3132)
    end if
  end subroutine check_ans_s
  subroutine check_ans(ans,msg)
    logical :: ans(:)
    character(len=*) :: msg
    if (.not.all(ans)) then
       print *, msg
       call fmpi_abort(3132)
    end if
  end subroutine check_ans
  subroutine test_integer_only_reductions(count)
    integer, intent(in) :: count

    integer, dimension(count) :: intin, intout, intsol
    integer rank, i, sz
    call fmpi_comm_rank(rank)
    call fmpi_comm_size(sz)

    ! test bor bitwise or
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_bor'
    if (sz < 3) then
       intsol = sz - 1
    else
       intsol = 3
    end if
    intin = mod(rank,4)
    intout = 0
    call fmpi_allreduce(intin, intout,  FMPI_bor)
    call check_ans(intout==intsol,'error for type mpi_integer and op mpi_bor')

    ! test band bitwise and 
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_band'
    ! see bottom for function definitions
    if (rank == 0) then
       intin = (/(i,i=1,count)/)
    else
       intin = -1
    end if
    intsol = (/(i,i=1,count)/)
    intout = 0
    call fmpi_allreduce( intin, intout,  FMPI_band)
    call check_ans(intout==intsol,'error for type mpi_integer and op mpi_band 1')
    if (rank == 0) then
       intin = 0
    else
       intin = (/(i,i=1,count)/)
    end if
    intsol = 0
    intout = 0
    call fmpi_allreduce( intin, intout, FMPI_band)
    call check_ans(intout==intsol,'error for type mpi_integer and op mpi_band 2')

    ! test bxor 
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_bxor'
    if (rank == 0) then
       intin =  213123
    else
       intin = 0
    end if
    intsol = 213123
    intout = 0
    call fmpi_allreduce( intin, intout, FMPI_bxor)
    call check_ans(intout==intsol,'error for type mpi_integer and op mpi_bxor 1')

    intin = 0
    intsol = 0
    intout = 0
    call fmpi_allreduce( intin, intout,  FMPI_bxor)
    call check_ans(intout==intsol,'error for type mpi_integer and op mpi_bxor 2')

    ! assumes -1 == all bits set
    intin = (-1)
    intsol = - mod(sz,2)
    intout = 0
    call fmpi_allreduce( intin, intout, FMPI_bxor)
    call check_ans(intout==intsol,'error for type mpi_integer and op mpi_bxor 3')
  end subroutine test_integer_only_reductions


  subroutine test_integer_reductions(count)
    integer, intent(in) :: count

    integer, dimension(count) :: datain, dataout, datasol
    integer rank
    integer sz,i
    call fmpi_comm_rank(rank)
    call fmpi_comm_size(sz)

    ! test sum 
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_sum'
    datain = (/(i,i=1,count)/)
    datasol = sz * datain
    dataout = 0
    call fmpi_allreduce(datain, dataout, FMPI_sum)
    call check_ans(dataout==datasol, 'error for type mpi_integer and op mpi_sum')

    ! test prod
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_prod'
    datain = (/(i,i=1,count)/)
    datasol = datain**sz
    dataout = 0
    call fmpi_allreduce(datain, dataout, FMPI_prod)    
    call check_ans(dataout==datasol,'error for type mpi_integer and op mpi_prod')

    ! test max
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_max'
    datain = rank + (/(i,i=1,count)/)
    datasol = sz - 1 + (/(i,i=1,count)/)
    dataout = 0
    call fmpi_allreduce(datain, dataout, FMPI_max)
    call check_ans(dataout==datasol,'error for type mpi_integer and op mpi_max')

    ! test min
    datain = rank + (/(i,i=1,count)/)
    datasol =  (/(i,i=1,count)/)
    dataout = 0
    call fmpi_allreduce(datain, dataout, FMPI_MIN)
    call check_ans(dataout==datasol,'error for type mpi_integer and op mpi_min')
  end subroutine test_integer_reductions
 subroutine test_s_integer_reductions
   integer :: datain, dataout, datasol
    integer rank
    integer sz
    call fmpi_comm_rank(rank)
    call fmpi_comm_size(sz)

    ! test sum 
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_sum'
    datain = 92
    datasol = sz * datain
    dataout = 0
    call fmpi_allreduce(datain, dataout, FMPI_sum)
    call check_ans_s(dataout==datasol, 'error for type mpi_integer and op mpi_sum')

    ! test prod
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_prod'
    datain = 1
    datasol = datain**sz
    dataout = 0
    call fmpi_allreduce(datain, dataout, Fmpi_prod)    
    call check_ans_s(dataout==datasol,'error for type mpi_integer and op mpi_prod')

    ! test max
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_max'
    datain = rank + 1
    datasol = sz 
    dataout = 0
    call fmpi_allreduce(datain, dataout, FMPI_max)
    call check_ans_s(dataout==datasol,'error for type mpi_integer and op mpi_max')

    ! test min
    datain = rank + 1
    datasol =  1
    dataout = 0
    call fmpi_allreduce(datain, dataout, FMPI_min)
    call check_ans_s(dataout==datasol,'error for type mpi_integer and op mpi_min')
  end subroutine test_s_integer_reductions

  subroutine test_logical_reductions(count)
    integer, intent(in) :: count

    logical, dimension(count) :: login, logout, logsol
    integer rank
    integer sz
    call fmpi_comm_rank(rank)
    call fmpi_comm_size(sz)

    ! test lor logical or
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_lor'

    login = (mod(rank,2) == 1)
    logsol = (sz > 1)
    logout = .false.
    call fmpi_allreduce(login, logout, FMPI_lor)
    call check_ans(logout.eqv.logsol,'error for type mpi_logical and op mpi_lor 1')
    login = .false.
    logsol = .false.
    logout = .false.
    call fmpi_allreduce( login, logout, FMPI_lor)
    call check_ans(logout.eqv.logsol,'error for type mpi_logical and op mpi_lor 2')

    ! test lxor 
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_lxor'
    login = (rank == 0)
    logsol = .true.
    logout = .false.
    call fmpi_allreduce( login, logout, Fmpi_lxor)
    call check_ans(logout.eqv.logsol, 'error for type mpi_logical and op mpi_lxor 1')
    login = .false.
    logsol = .false.
    logout = .false.
    call fmpi_allreduce( login, logout, Fmpi_lxor)
    call check_ans(logout.eqv.logsol, 'error for type mpi_logical and op mpi_lxor 2')
    login = .true.
    logsol = (mod(sz,2) == 1)
    logout = .false.
    call fmpi_allreduce( login, logout, Fmpi_lxor)
    call check_ans(logout.eqv.logsol, 'error for type mpi_logical and op mpi_lxor 3')

    ! test land  bitwise and
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_land'
    login = (mod(rank,2) .eq. 1)
    logsol = .false.
    logout = .false.
    call fmpi_allreduce( login, logout,  Fmpi_land)
    call check_ans(logout.eqv.logsol,'error for type mpi_logical and op mpi_land 1')

    login = .true.
    logsol = .true.
    logout = .false.
    call fmpi_allreduce( login, logout,  Fmpi_land)
    call check_ans(logout.eqv.logsol,'error for type mpi_logical and op mpi_land 2')
  end subroutine test_logical_reductions
 subroutine test_real_reductions(count)
    integer, intent(in) :: count

    real, dimension(count) :: datain, dataout, datasol
    integer rank
    integer sz,i
    call fmpi_comm_rank(rank)
    call fmpi_comm_size(sz)

    ! test sum 
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_sum'
    datain = (/(i,i=1,count)/)
    datasol = sz * datain
    dataout = 0
    call fmpi_allreduce(datain, dataout, Fmpi_sum)
    call check_ans(dataout==datasol, 'error for type mpi_integer and op mpi_sum')

    ! test prod
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_prod'
    datain = (/(i,i=1,count)/)
    datasol = datain**sz
    dataout = 0
    call fmpi_allreduce(datain, dataout, Fmpi_prod)    
    call check_ans(dataout==datasol,'error for type mpi_integer and op mpi_prod')

    ! test max
    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_max'
    datain = rank + (/(i,i=1,count)/)
    datasol = sz - 1 + (/(i,i=1,count)/)
    dataout = 0
    call fmpi_allreduce(datain, dataout, Fmpi_max)
    call check_ans(dataout==datasol,'error for type mpi_integer and op mpi_max')

    ! test min
    datain = rank + (/(i,i=1,count)/)
    datasol =  (/(i,i=1,count)/)
    dataout = 0
    call fmpi_allreduce(datain, dataout, Fmpi_min)
    call check_ans(dataout==datasol,'error for type mpi_integer and op mpi_min')
  end subroutine test_real_reductions

end program allreduce
!!$
!!$    ! test maxloc 
!!$    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_maxloc'
!!$    do  i=1, count
!!$       intin(2*i-1) = (rank + i)
!!$       intin(2*i)   = rank
!!$       intsol(2*i-1) = (count - 1 + i)
!!$       intsol(2*i) = (count-1)
!!$       intout(2*i-1) = 0
!!$       intout(2*i)   = 0
!!$    end do
!!$    call fmpi_allreduce( intin, intout, count,       Fmpi_2integer, Fmpi_maxloc)
!!$    do  i=1, count
!!$       if (intout(2*i-1) .ne. intsol(2*i-1) .or.    intout(2*i) .ne. intsol(2*i)) then
!!$          errcnt = errcnt + 1
!!$          fnderr = .true. 
!!$       endif
!!$    end do
!!$    if (fnderr) then
!!$       print *, 'error for type mpi_2integer and op mpi_maxloc'
!!$    endif
!!$
!!$
!!$    fnderr = .false.
!!$    do  i=1, count
!!$       realin(2*i-1) = (rank + i)
!!$       realin(2*i)   = rank
!!$       realsol(2*i-1) = (count - 1 + i)
!!$       realsol(2*i) = (count-1)
!!$       realout(2*i-1) = 0
!!$       realout(2*i)   = 0
!!$    end do
!!$    call fmpi_allreduce( realin, realout, count,       Fmpi_2real, Fmpi_maxloc)
!!$    do  i=1, count
!!$       if (realout(2*i-1) .ne. realsol(2*i-1) .or.realout(2*i) .ne. realsol(2*i)) then
!!$
!!$          errcnt = errcnt + 1
!!$          fnderr = .true. 
!!$       endif
!!$    end do
!!$    if (fnderr) then
!!$       print *, 'error for type mpi_2real and op mpi_maxloc'
!!$    endif
!!$
!!$
!!$    fnderr = .false.
!!$    do  i=1, count
!!$       dblein(2*i-1) = (rank + i)
!!$       dblein(2*i)   = rank
!!$       dblesol(2*i-1) = (count - 1 + i)
!!$       dblesol(2*i) = (count-1)
!!$       dbleout(2*i-1) = 0
!!$       dbleout(2*i)   = 0
!!$    end do
!!$    call fmpi_allreduce( dblein, dbleout, count,       mpi_2double_precision, Fmpi_maxloc)
!!$    do  i=1, count
!!$       if (dbleout(2*i-1) .ne. dblesol(2*i-1) .or.  dbleout(2*i) .ne. dblesol(2*i)) then
!!$
!!$          errcnt = errcnt + 1
!!$          fnderr = .true. 
!!$       endif
!!$    end do
!!$    if (fnderr) then
!!$       print *, 'error for type mpi_2double_precision and op mpi_maxloc'
!!$    endif
!!$
!!$
!!$    if (errcnt .gt. 0) then
!!$       print *, 'found ', errcnt, ' errors on ', rank,             ' for mpi_maxloc'
!!$    endif
!!$    errcnt = 0
!!$
!!$    ! test minloc 
!!$    if (fmpi_is_root(rank)) write(*,'(a)',advance='no') ' mpi_minloc'
!!$
!!$
!!$    fnderr = .false.
!!$    do  i=1, count
!!$       intin(2*i-1) = (rank + i)
!!$       intin(2*i)   = rank
!!$       intsol(2*i-1) = i
!!$       intsol(2*i) = 0
!!$       intout(2*i-1) = 0
!!$       intout(2*i)   = 0
!!$    end do
!!$    call fmpi_allreduce( intin, intout, count,       Fmpi_2integer, Fmpi_minloc)
!!$    do  i=1, count
!!$       if (intout(2*i-1) .ne. intsol(2*i-1) .or.   intout(2*i) .ne. intsol(2*i)) then
!!$
!!$          errcnt = errcnt + 1
!!$          fnderr = .true. 
!!$       endif
!!$    end do
!!$    if (fnderr) then
!!$       print *, 'error for type mpi_2integer and op mpi_minloc'
!!$    endif
!!$
!!$
!!$    fnderr = .false.
!!$    do  i=1, count
!!$       realin(2*i-1) = (rank + i)
!!$       realin(2*i)   = rank
!!$       realsol(2*i-1) = i
!!$       realsol(2*i) = 0
!!$       realout(2*i-1) = 0
!!$       realout(2*i)   = 0
!!$    end do
!!$    call fmpi_allreduce( realin, realout, count,       Fmpi_2real, Fmpi_minloc)
!!$    do  i=1, count
!!$       if (realout(2*i-1) .ne. realsol(2*i-1) .or.  realout(2*i) .ne. realsol(2*i)) then
!!$          errcnt = errcnt + 1
!!$          fnderr = .true. 
!!$       endif
!!$    end do
!!$    if (fnderr) then
!!$       print *, 'error for type mpi_2real and op mpi_minloc'
!!$    endif
!!$
!!$
!!$    fnderr = .false.
!!$    do  i=1, count
!!$       dblein(2*i-1) = (rank + i)
!!$       dblein(2*i)   = rank
!!$       dblesol(2*i-1) = i
!!$       dblesol(2*i) = 0
!!$       dbleout(2*i-1) = 0
!!$       dbleout(2*i)   = 0
!!$    end do
!!$    call fmpi_allreduce( dblein, dbleout, count,       mpi_2double_precision, Fmpi_minloc)yy
!!$    do  i=1, count
!!$       if (dbleout(2*i-1) .ne. dblesol(2*i-1) .or.  dbleout(2*i) .ne. dblesol(2*i)) then
!!$          errcnt = errcnt + 1
!!$          fnderr = .true. 
!!$       endif
!!$    end do
!!$    if (fnderr) then
!!$       print *, 'error for type mpi_2double_precision and op mpi_minloc'
!!$    endif
!!$
!!$
!!$    if (errcnt .gt. 0) then
!!$       print *, 'found ', errcnt, ' errors on ', rank,             ' for mpi_minloc'
!!$    endif
!!$    errcnt = 0
