! This program checks the mpi functions for collective communications
! It can run on any number of processes
! It uses allocatable arrays and thus provides a check on the f90 interface
program collective
  use mpi
  implicit none
  ! Change this type below to test other datatype 
  real,allocatable::sendbuf(:),recvbuf(:),ans(:),data(:),sendmat(:,:),recvmat(:,:),x(:)
  real, pointer :: point(:)

  integer :: sz, i, j
  integer, parameter :: length = 31 ! length for array test
  integer, allocatable :: recvcounts(:), sendcounts(:), rdispls(:), sdispls(:)
  integer, allocatable :: rankl(:), rev(:)
  integer n
  integer rank

  call fmpi_init
  call fmpi_comm_rank(rank)
  call fmpi_comm_size(sz)
  allocate(rev(sz), rankl(sz),data(length*sz))
  data = (/((sz**2*i+j,i=0,length-1),j=0,sz-1)/)
  rankl = (/(i,i=0,sz-1)/)  ! list of ranks
  rev = (/(i,i=sz,1,-1)/)
  ! array  collectives to and from root
  allocate(sendbuf(length))
  if (fmpi_is_root(rank)) then 
     allocate(ans(sz*length),recvbuf(sz*length))
  else 
     allocate(ans(0),recvbuf(0))
  end if

  if (fmpi_is_root(rank)) print *, 'checking array bcast'
  sendbuf = rank+data(1:length)
  call fmpi_bcast(sendbuf)
  call check_msg(all(sendbuf==data(1:length)), 'bcast a',recvbuf)

  if (fmpi_is_root(rank)) write(*,'(a)',advance='no')' array gather'
  sendbuf = rank+data(1:length)
  if (fmpi_is_root(rank)) ans = data
  recvbuf = -1
  call fmpi_gather(sendbuf, recvbuf)
  call check_msg(all(recvbuf==ans), 'gather a',recvbuf)
  deallocate(ans,recvbuf,sendbuf)

  ! Now scatter each node its rank and some data
  if (fmpi_is_root(rank)) then
     write(*,'(a)',advance='no')' array scatter'
     allocate(sendbuf(sz*length))
     sendbuf = data
  else
     allocate(sendbuf(0))
  end if
  allocate(recvbuf(length),ans(length))
  recvbuf = -1
  ans = rank+data(1:length)
  call fmpi_scatter(sendbuf, recvbuf)
  call check_msg(all(ans==recvbuf), 'scatter a',sendbuf(1:1))
  deallocate(sendbuf,recvbuf,ans)


  if (fmpi_is_root(rank)) write(*,'(a)',advance='no')' array allgather'
  allocate(sendbuf(length),recvbuf(sz*length))
  sendbuf = rank+data(1:length)
  recvbuf = -1
  call fmpi_allgather(sendbuf, recvbuf)
  call check_msg(all(recvbuf==data), 'allgather a',recvbuf)
  deallocate(sendbuf,recvbuf)


  n = (sz*(sz-1))/2
  if (fmpi_is_root(rank)) then
     write(*,'(a)',advance='no')' array gatherv'
     allocate(recvbuf(n),ans(n),rdispls(sz),recvcounts(sz))
     ans = (/((sz*i+j,j=1,i),i=0,sz-1)/)
     recvcounts = rankl    
     rdispls = (/((i*(i-1))/2,i=0,sz-1)/)
  else
     allocate(recvbuf(0),ans(0),rdispls(0),recvcounts(0))
  end if
  allocate(sendbuf(rank))
  sendbuf = (/(sz*rank+i,i=1,rank)/)
  recvbuf = -1
  call fmpi_gatherv(sendbuf, recvbuf, recvcounts, rdispls)
  call check_msg(all(recvbuf==ans), 'gatherv a',recvbuf)

  n = (sz*(sz-1))/2
  if (fmpi_is_root(rank)) then
     write(*,'(a)',advance='no')' array gathera'
     ans = (/((sz*i+j,j=1,i),i=0,sz-1)/)
  end if
  sendbuf = (/(sz*rank+i,i=1,rank)/)
  call fmpi_gathera(sendbuf, point)
  if (fmpi_is_root(rank)) then
     call check_msg(all(point==ans), 'gathera a',recvbuf)
  else
     call check_msg(size(point)==0, 'gathera pointer size incorrect a',recvbuf)
  end if
  deallocate(point)
     
  ! Now scatter the data back to each node
  if (fmpi_is_root(rank)) write(*,'(a)',advance='no')' array scatterv'  
  sendbuf = -1
  call fmpi_scatterv(recvbuf, recvcounts, rdispls, sendbuf)
  call check_msg(all((/(sz*rank+i,i=1,rank)/)==sendbuf), 'scatterv a',sendbuf)
  deallocate(recvbuf,ans,rdispls,recvcounts)



  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' array allgatherv'
  allocate(recvbuf(n),ans(n),rdispls(sz),recvcounts(sz))
  ans = (/((sz*i+j,j=1,i),i=0,sz-1)/)
  recvcounts = rankl    
  rdispls = (/((i*(i-1))/2,i=0,sz-1)/)
  recvbuf = -1
  call fmpi_allgatherv(sendbuf, recvbuf, recvcounts, rdispls)
  call check_msg(all(recvbuf==ans), 'allgatherv a',recvbuf)
  deallocate(recvbuf,ans,rdispls,recvcounts,data,sendbuf)



  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' array alltoall'
  ! Every processor sends its rank*sz + the rank of the receiving process
  ! alltoall is most naturally overloaded with matrix argument
  n = sz*length
  allocate(sendbuf(n),recvbuf(n),ans(n),data(n),sendmat(length,sz),recvmat(length,sz))
  data=(/((j+length*i,j=0,length-1),i=0,sz-1)/)
  sendbuf = rank*n+data
  ans = rank*length+(/((j+n*i,j=0,length-1),i=0,sz-1)/)
  recvbuf = -1
  sendmat = reshape(sendbuf,(/length,sz/))
  call fmpi_alltoall(sendmat,recvmat)
  recvbuf = reshape(recvmat,(/n/))
  call check_msg(all(recvbuf==ans), 'alltoall a',recvbuf)
  deallocate(sendbuf,recvbuf,ans,data,sendmat,recvmat)

  ! send in reverse order and receieve in reverse order
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' array alltoallv'
  ! Fully randomised test each process sends a random amount of data to each process
  ! This is then sent back and compared
  ! Make sure that each process is generating different random numbers
  call random_seed(size=n)               ! how long is the seed
  allocate(x(n))
  call random_number(x)
  x = x*real(huge(n))/real(sz)*real(rank)
  call system_clock(n)
  x(1)=real(n)/(1+real(rank))
  call random_seed(put=int(x)) ! Set it to a function of the rank
  deallocate(x)

  allocate(x(sz),sendcounts(sz),recvcounts(sz),sdispls(sz),rdispls(sz))
  call random_number(x)
  sendcounts=floor(length*x)          ! how much data to send
  n = sum(sendcounts)
  allocate(sendbuf(n),ans(n))  ! allocate the space
  call random_number(sendbuf)         ! generate the data
  ans = sendbuf                       ! remember the data
  call fmpi_alltoall(sendcounts,recvcounts) ! Send a count to each processor
  allocate(recvbuf(sum(recvcounts)))  ! allocate the receive buffer
  sdispls(1)=0                        ! Build send and recieve displacements
  rdispls(1)=0
  do i=1,sz-1
     sdispls(i+1) =  sdispls(i)+sendcounts(i)
     rdispls(i+1) =  rdispls(i)+recvcounts(i)
  end do
  recvbuf = -1
  call fmpi_alltoallv(sendbuf, sendcounts, sdispls, recvbuf, recvcounts, rdispls)
  sendbuf = -1
  call fmpi_alltoallv(recvbuf, recvcounts, rdispls, sendbuf, sendcounts, sdispls)
  call check_msg(all(sendbuf==ans), 'alltoallv a',recvbuf)
  deallocate(recvbuf,sendbuf,ans)

  ! Now repeat all the functions when the lengths are one
  ! Gather a scalar collectives to and from root

  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar bcast'
  allocate(recvbuf(sz), ans(sz), sendbuf(1))
  sendbuf = rank
  call fmpi_bcast(sendbuf(1))
  call check_msg(sendbuf(1)==0, 'bcast s',recvbuf)
  sendbuf = rank
  ans = rankl
  if (.not. fmpi_is_root(rank)) ans = -1
  recvbuf = -1
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar gather'
  call fmpi_gather(sendbuf(1), recvbuf)
  call check_msg(all(recvbuf==ans), 'gather s',recvbuf)
  ans = ans(rev) ! reverse the answer
  rdispls = rev-1
  recvbuf = -1
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar gatherv'
  call fmpi_gatherv(sendbuf(1), recvbuf, rdispls)
  call check_msg(all(recvbuf==ans), 'gatherv s',recvbuf)
  ! No scatter each node its rank
  recvbuf = rankl 
  sendbuf(1) = -1
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar scatter'
  call fmpi_scatter(recvbuf, sendbuf(1))
  call check_msg(rank==sendbuf(1), 'scatter s',sendbuf(1:1))
  ! No scatter each node its rank but from a backwards list
  recvbuf = rankl(rev) ! backwards
  sendbuf(1) = -1
  rdispls = rev-1
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar scatterv'
  call fmpi_scatterv(recvbuf, rdispls, sendbuf(1))
  call check_msg(rank==sendbuf(1), 'scatterv s',sendbuf(1:1))
  sendbuf = rank
  ans = rankl
  recvbuf = -1
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar allgather'
  call fmpi_allgather(sendbuf(1), recvbuf)
  call check_msg(all(recvbuf==ans), 'allgather s',recvbuf)
  ans = ans(rev) ! reverse the answer
  rdispls = rev-1
  recvbuf = -1
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar allgatherv'
  call fmpi_allgatherv(sendbuf(1), recvbuf, rdispls)
  call check_msg(all(recvbuf==ans), 'allgatherv s',recvbuf)
  ! Every processor sends its rank*sz + the rank of the receiving process
  deallocate(sendbuf)
  allocate(sendbuf(sz))
  sendbuf = rankl+sz*rank
  ans = rankl*sz+rank
  recvbuf = -1
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar alltoall'
  call fmpi_alltoall(sendbuf, recvbuf)
  call check_msg(all(recvbuf==ans), 'alltoall s',recvbuf)
  ! send in reverse order and receieve in reverse order
  recvbuf = -1
  rdispls = rev -1
  ans = (sz-rankl-1)*sz+ (sz-rank-1) 
  if (fmpi_is_root(rank))   write(*,'(a)',advance='no')' scalar alltoallv'
  call fmpi_alltoallv(sendbuf, rdispls, recvbuf, rdispls)
  call check_msg(all(recvbuf==ans), 'alltoallv s',recvbuf)
  deallocate(recvbuf, ans, sendbuf)
  if (fmpi_is_root(rank)) print *,'Collective operations all passed on',sz,'processes'
  call fmpi_finalize
contains
  subroutine check_msg(f,s,d)
    logical, intent(in) :: f
    character(len=*) :: s
    real, optional :: d(:)
    integer rank
    call fmpi_barrier
    call fmpi_comm_rank(rank)
    if (f) return
    write(*,*)s, ' failed on node',rank  
    if (present(d)) print *,d
    call fmpi_abort(31321)
  end subroutine check_msg

end program collective


